# -*- coding: utf-8 -*-
import platform
from riak.transports.feature_detect import FeatureDetection

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest


class IncompleteTransport(FeatureDetection):
    pass


class DummyTransport(FeatureDetection):
    def __init__(self, version):
        self._version = version

    def _server_version(self):
        return self._version


class FeatureDetectionTest(unittest.TestCase):
    def test_implements_server_version(self):
        t = IncompleteTransport()

        with self.assertRaises(NotImplementedError):
            t.server_version

    def test_pre_10(self):
        t = DummyTransport("0.14.2")
        self.assertFalse(t.phaseless_mapred())
        self.assertFalse(t.pb_indexes())
        self.assertFalse(t.pb_search())
        self.assertFalse(t.pb_conditionals())
        self.assertFalse(t.quorum_controls())
        self.assertFalse(t.tombstone_vclocks())
        self.assertFalse(t.pb_head())
        self.assertFalse(t.pb_clear_bucket_props())
        self.assertFalse(t.pb_all_bucket_props())
        self.assertFalse(t.counters())
        self.assertFalse(t.stream_indexes())
        self.assertFalse(t.index_term_regex())
        self.assertFalse(t.bucket_types())
        self.assertFalse(t.datatypes())
        self.assertFalse(t.preflists())
        self.assertFalse(t.write_once())

    def test_10(self):
        t = DummyTransport("1.0.3")
        self.assertFalse(t.phaseless_mapred())
        self.assertFalse(t.pb_indexes())
        self.assertFalse(t.pb_search())
        self.assertTrue(t.pb_conditionals())
        self.assertTrue(t.quorum_controls())
        self.assertTrue(t.tombstone_vclocks())
        self.assertTrue(t.pb_head())
        self.assertFalse(t.pb_clear_bucket_props())
        self.assertFalse(t.pb_all_bucket_props())
        self.assertFalse(t.counters())
        self.assertFalse(t.stream_indexes())
        self.assertFalse(t.index_term_regex())
        self.assertFalse(t.bucket_types())
        self.assertFalse(t.datatypes())
        self.assertFalse(t.preflists())
        self.assertFalse(t.write_once())

    def test_11(self):
        t = DummyTransport("1.1.4")
        self.assertTrue(t.phaseless_mapred())
        self.assertFalse(t.pb_indexes())
        self.assertFalse(t.pb_search())
        self.assertTrue(t.pb_conditionals())
        self.assertTrue(t.quorum_controls())
        self.assertTrue(t.tombstone_vclocks())
        self.assertTrue(t.pb_head())
        self.assertFalse(t.pb_clear_bucket_props())
        self.assertFalse(t.pb_all_bucket_props())
        self.assertFalse(t.counters())
        self.assertFalse(t.stream_indexes())
        self.assertFalse(t.index_term_regex())
        self.assertFalse(t.bucket_types())
        self.assertFalse(t.datatypes())
        self.assertFalse(t.preflists())
        self.assertFalse(t.write_once())

    def test_12(self):
        t = DummyTransport("1.2.0")
        self.assertTrue(t.phaseless_mapred())
        self.assertTrue(t.pb_indexes())
        self.assertTrue(t.pb_search())
        self.assertTrue(t.pb_conditionals())
        self.assertTrue(t.quorum_controls())
        self.assertTrue(t.tombstone_vclocks())
        self.assertTrue(t.pb_head())
        self.assertFalse(t.pb_clear_bucket_props())
        self.assertFalse(t.pb_all_bucket_props())
        self.assertFalse(t.counters())
        self.assertFalse(t.stream_indexes())
        self.assertFalse(t.index_term_regex())
        self.assertFalse(t.bucket_types())
        self.assertFalse(t.datatypes())
        self.assertFalse(t.preflists())
        self.assertFalse(t.write_once())

    def test_12_loose(self):
        t = DummyTransport("1.2.1p3")
        self.assertTrue(t.phaseless_mapred())
        self.assertTrue(t.pb_indexes())
        self.assertTrue(t.pb_search())
        self.assertTrue(t.pb_conditionals())
        self.assertTrue(t.quorum_controls())
        self.assertTrue(t.tombstone_vclocks())
        self.assertTrue(t.pb_head())
        self.assertFalse(t.pb_clear_bucket_props())
        self.assertFalse(t.pb_all_bucket_props())
        self.assertFalse(t.counters())
        self.assertFalse(t.stream_indexes())
        self.assertFalse(t.index_term_regex())
        self.assertFalse(t.bucket_types())
        self.assertFalse(t.datatypes())
        self.assertFalse(t.preflists())
        self.assertFalse(t.write_once())

    def test_14(self):
        t = DummyTransport("1.4.0rc1")
        self.assertTrue(t.phaseless_mapred())
        self.assertTrue(t.pb_indexes())
        self.assertTrue(t.pb_search())
        self.assertTrue(t.pb_conditionals())
        self.assertTrue(t.quorum_controls())
        self.assertTrue(t.tombstone_vclocks())
        self.assertTrue(t.pb_head())
        self.assertTrue(t.pb_clear_bucket_props())
        self.assertTrue(t.pb_all_bucket_props())
        self.assertTrue(t.counters())
        self.assertTrue(t.stream_indexes())
        self.assertFalse(t.index_term_regex())
        self.assertFalse(t.bucket_types())
        self.assertFalse(t.datatypes())
        self.assertFalse(t.preflists())
        self.assertFalse(t.write_once())

    def test_144(self):
        t = DummyTransport("1.4.6")
        self.assertTrue(t.phaseless_mapred())
        self.assertTrue(t.pb_indexes())
        self.assertTrue(t.pb_search())
        self.assertTrue(t.pb_conditionals())
        self.assertTrue(t.quorum_controls())
        self.assertTrue(t.tombstone_vclocks())
        self.assertTrue(t.pb_head())
        self.assertTrue(t.pb_clear_bucket_props())
        self.assertTrue(t.pb_all_bucket_props())
        self.assertTrue(t.counters())
        self.assertTrue(t.stream_indexes())
        self.assertTrue(t.index_term_regex())
        self.assertFalse(t.bucket_types())
        self.assertFalse(t.datatypes())
        self.assertFalse(t.preflists())
        self.assertFalse(t.write_once())

    def test_20(self):
        t = DummyTransport("2.0.1")
        self.assertTrue(t.phaseless_mapred())
        self.assertTrue(t.pb_indexes())
        self.assertTrue(t.pb_search())
        self.assertTrue(t.pb_conditionals())
        self.assertTrue(t.quorum_controls())
        self.assertTrue(t.tombstone_vclocks())
        self.assertTrue(t.pb_head())
        self.assertTrue(t.pb_clear_bucket_props())
        self.assertTrue(t.pb_all_bucket_props())
        self.assertTrue(t.counters())
        self.assertTrue(t.stream_indexes())
        self.assertTrue(t.index_term_regex())
        self.assertTrue(t.bucket_types())
        self.assertTrue(t.datatypes())
        self.assertFalse(t.preflists())
        self.assertFalse(t.write_once())

    def test_21(self):
        t = DummyTransport("2.1.0")
        self.assertTrue(t.phaseless_mapred())
        self.assertTrue(t.pb_indexes())
        self.assertTrue(t.pb_search())
        self.assertTrue(t.pb_conditionals())
        self.assertTrue(t.quorum_controls())
        self.assertTrue(t.tombstone_vclocks())
        self.assertTrue(t.pb_head())
        self.assertTrue(t.pb_clear_bucket_props())
        self.assertTrue(t.pb_all_bucket_props())
        self.assertTrue(t.counters())
        self.assertTrue(t.stream_indexes())
        self.assertTrue(t.index_term_regex())
        self.assertTrue(t.bucket_types())
        self.assertTrue(t.datatypes())
        self.assertTrue(t.preflists())
        self.assertTrue(t.write_once())

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import platform
from riak.mapreduce import RiakKeyFilter
from riak import key_filter

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest


class FilterTests(unittest.TestCase):
    def test_simple(self):
        f1 = RiakKeyFilter("tokenize", "-", 1)
        self.assertEqual(f1._filters, [["tokenize", "-", 1]])

    def test_add(self):
        f1 = RiakKeyFilter("tokenize", "-", 1)
        f2 = RiakKeyFilter("eq", "2005")
        f3 = f1 + f2
        self.assertEqual(list(f3), [["tokenize", "-", 1], ["eq", "2005"]])

    def test_and(self):
        f1 = RiakKeyFilter("starts_with", "2005-")
        f2 = RiakKeyFilter("ends_with", "-01")
        f3 = f1 & f2
        self.assertEqual(list(f3),
                         [["and",
                           [["starts_with", "2005-"]],
                           [["ends_with", "-01"]]]])

    def test_multi_and(self):
        f1 = RiakKeyFilter("starts_with", "2005-")
        f2 = RiakKeyFilter("ends_with", "-01")
        f3 = RiakKeyFilter("matches", "-11-")
        f4 = f1 & f2 & f3
        self.assertEqual(list(f4), [["and",
                                     [["starts_with", "2005-"]],
                                     [["ends_with", "-01"]],
                                     [["matches", "-11-"]],
                                     ]])

    def test_or(self):
        f1 = RiakKeyFilter("starts_with", "2005-")
        f2 = RiakKeyFilter("ends_with", "-01")
        f3 = f1 | f2
        self.assertEqual(list(f3), [["or", [["starts_with", "2005-"]],
                                     [["ends_with", "-01"]]]])

    def test_multi_or(self):
        f1 = RiakKeyFilter("starts_with", "2005-")
        f2 = RiakKeyFilter("ends_with", "-01")
        f3 = RiakKeyFilter("matches", "-11-")
        f4 = f1 | f2 | f3
        self.assertEqual(list(f4), [["or",
                                     [["starts_with", "2005-"]],
                                     [["ends_with", "-01"]],
                                     [["matches", "-11-"]],
                                     ]])

    def test_chaining(self):
        f1 = key_filter.tokenize("-", 1).eq("2005")
        f2 = key_filter.tokenize("-", 2).eq("05")
        f3 = f1 & f2
        self.assertEqual(list(f3), [["and",
                                     [["tokenize", "-", 1], ["eq", "2005"]],
                                     [["tokenize", "-", 2], ["eq", "05"]]
                                     ]])



from six import *
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: riak.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='riak.proto',
  package='',
  serialized_pb='\n\nriak.proto\"/\n\x0cRpbErrorResp\x12\x0e\n\x06\x65rrmsg\x18\x01 \x02(\x0c\x12\x0f\n\x07\x65rrcode\x18\x02 \x02(\r\"<\n\x14RpbGetServerInfoResp\x12\x0c\n\x04node\x18\x01 \x01(\x0c\x12\x16\n\x0eserver_version\x18\x02 \x01(\x0c\"%\n\x07RpbPair\x12\x0b\n\x03key\x18\x01 \x02(\x0c\x12\r\n\x05value\x18\x02 \x01(\x0c\"/\n\x0fRpbGetBucketReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\x12\x0c\n\x04type\x18\x02 \x01(\x0c\"2\n\x10RpbGetBucketResp\x12\x1e\n\x05props\x18\x01 \x02(\x0b\x32\x0f.RpbBucketProps\"O\n\x0fRpbSetBucketReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\x12\x1e\n\x05props\x18\x02 \x02(\x0b\x32\x0f.RpbBucketProps\x12\x0c\n\x04type\x18\x03 \x01(\x0c\"1\n\x11RpbResetBucketReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\x12\x0c\n\x04type\x18\x02 \x01(\x0c\"#\n\x13RpbGetBucketTypeReq\x12\x0c\n\x04type\x18\x01 \x02(\x0c\"C\n\x13RpbSetBucketTypeReq\x12\x0c\n\x04type\x18\x01 \x02(\x0c\x12\x1e\n\x05props\x18\x02 \x02(\x0b\x32\x0f.RpbBucketProps\"-\n\tRpbModFun\x12\x0e\n\x06module\x18\x01 \x02(\x0c\x12\x10\n\x08\x66unction\x18\x02 \x02(\x0c\"9\n\rRpbCommitHook\x12\x1a\n\x06modfun\x18\x01 \x01(\x0b\x32\n.RpbModFun\x12\x0c\n\x04name\x18\x02 \x01(\x0c\"\xb0\x05\n\x0eRpbBucketProps\x12\r\n\x05n_val\x18\x01 \x01(\r\x12\x12\n\nallow_mult\x18\x02 \x01(\x08\x12\x17\n\x0flast_write_wins\x18\x03 \x01(\x08\x12!\n\tprecommit\x18\x04 \x03(\x0b\x32\x0e.RpbCommitHook\x12\x1c\n\rhas_precommit\x18\x05 \x01(\x08:\x05\x66\x61lse\x12\"\n\npostcommit\x18\x06 \x03(\x0b\x32\x0e.RpbCommitHook\x12\x1d\n\x0ehas_postcommit\x18\x07 \x01(\x08:\x05\x66\x61lse\x12 \n\x0c\x63hash_keyfun\x18\x08 \x01(\x0b\x32\n.RpbModFun\x12\x1b\n\x07linkfun\x18\t \x01(\x0b\x32\n.RpbModFun\x12\x12\n\nold_vclock\x18\n \x01(\r\x12\x14\n\x0cyoung_vclock\x18\x0b \x01(\r\x12\x12\n\nbig_vclock\x18\x0c \x01(\r\x12\x14\n\x0csmall_vclock\x18\r \x01(\r\x12\n\n\x02pr\x18\x0e \x01(\r\x12\t\n\x01r\x18\x0f \x01(\r\x12\t\n\x01w\x18\x10 \x01(\r\x12\n\n\x02pw\x18\x11 \x01(\r\x12\n\n\x02\x64w\x18\x12 \x01(\r\x12\n\n\x02rw\x18\x13 \x01(\r\x12\x14\n\x0c\x62\x61sic_quorum\x18\x14 \x01(\x08\x12\x13\n\x0bnotfound_ok\x18\x15 \x01(\x08\x12\x0f\n\x07\x62\x61\x63kend\x18\x16 \x01(\x0c\x12\x0e\n\x06search\x18\x17 \x01(\x08\x12)\n\x04repl\x18\x18 \x01(\x0e\x32\x1b.RpbBucketProps.RpbReplMode\x12\x14\n\x0csearch_index\x18\x19 \x01(\x0c\x12\x10\n\x08\x64\x61tatype\x18\x1a \x01(\x0c\x12\x12\n\nconsistent\x18\x1b \x01(\x08\x12\x12\n\nwrite_once\x18\x1c \x01(\x08\">\n\x0bRpbReplMode\x12\t\n\x05\x46\x41LSE\x10\x00\x12\x0c\n\x08REALTIME\x10\x01\x12\x0c\n\x08\x46ULLSYNC\x10\x02\x12\x08\n\x04TRUE\x10\x03\",\n\nRpbAuthReq\x12\x0c\n\x04user\x18\x01 \x02(\x0c\x12\x10\n\x08password\x18\x02 \x02(\x0c\"*\n\x14RpbToggleEncodingReq\x12\x12\n\nuse_native\x18\x01 \x02(\x08\"+\n\x15RpbToggleEncodingResp\x12\x12\n\nuse_native\x18\x01 \x02(\x08\x42!\n\x17\x63om.basho.riak.protobufB\x06RiakPB')



_RPBBUCKETPROPS_RPBREPLMODE = _descriptor.EnumDescriptor(
  name='RpbReplMode',
  full_name='RpbBucketProps.RpbReplMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FALSE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REALTIME', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FULLSYNC', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRUE', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1236,
  serialized_end=1298,
)


_RPBERRORRESP = _descriptor.Descriptor(
  name='RpbErrorResp',
  full_name='RpbErrorResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='errmsg', full_name='RpbErrorResp.errmsg', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='errcode', full_name='RpbErrorResp.errcode', index=1,
      number=2, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=14,
  serialized_end=61,
)


_RPBGETSERVERINFORESP = _descriptor.Descriptor(
  name='RpbGetServerInfoResp',
  full_name='RpbGetServerInfoResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node', full_name='RpbGetServerInfoResp.node', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='server_version', full_name='RpbGetServerInfoResp.server_version', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=63,
  serialized_end=123,
)


_RPBPAIR = _descriptor.Descriptor(
  name='RpbPair',
  full_name='RpbPair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='RpbPair.key', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='RpbPair.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=125,
  serialized_end=162,
)


_RPBGETBUCKETREQ = _descriptor.Descriptor(
  name='RpbGetBucketReq',
  full_name='RpbGetBucketReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bucket', full_name='RpbGetBucketReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='RpbGetBucketReq.type', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=164,
  serialized_end=211,
)


_RPBGETBUCKETRESP = _descriptor.Descriptor(
  name='RpbGetBucketResp',
  full_name='RpbGetBucketResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='props', full_name='RpbGetBucketResp.props', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=213,
  serialized_end=263,
)


_RPBSETBUCKETREQ = _descriptor.Descriptor(
  name='RpbSetBucketReq',
  full_name='RpbSetBucketReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bucket', full_name='RpbSetBucketReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='props', full_name='RpbSetBucketReq.props', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='RpbSetBucketReq.type', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=265,
  serialized_end=344,
)


_RPBRESETBUCKETREQ = _descriptor.Descriptor(
  name='RpbResetBucketReq',
  full_name='RpbResetBucketReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bucket', full_name='RpbResetBucketReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='RpbResetBucketReq.type', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=346,
  serialized_end=395,
)


_RPBGETBUCKETTYPEREQ = _descriptor.Descriptor(
  name='RpbGetBucketTypeReq',
  full_name='RpbGetBucketTypeReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='RpbGetBucketTypeReq.type', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=397,
  serialized_end=432,
)


_RPBSETBUCKETTYPEREQ = _descriptor.Descriptor(
  name='RpbSetBucketTypeReq',
  full_name='RpbSetBucketTypeReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='RpbSetBucketTypeReq.type', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='props', full_name='RpbSetBucketTypeReq.props', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=434,
  serialized_end=501,
)


_RPBMODFUN = _descriptor.Descriptor(
  name='RpbModFun',
  full_name='RpbModFun',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='module', full_name='RpbModFun.module', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='function', full_name='RpbModFun.function', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=503,
  serialized_end=548,
)


_RPBCOMMITHOOK = _descriptor.Descriptor(
  name='RpbCommitHook',
  full_name='RpbCommitHook',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='modfun', full_name='RpbCommitHook.modfun', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='RpbCommitHook.name', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=550,
  serialized_end=607,
)


_RPBBUCKETPROPS = _descriptor.Descriptor(
  name='RpbBucketProps',
  full_name='RpbBucketProps',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='n_val', full_name='RpbBucketProps.n_val', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='allow_mult', full_name='RpbBucketProps.allow_mult', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='last_write_wins', full_name='RpbBucketProps.last_write_wins', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='precommit', full_name='RpbBucketProps.precommit', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='has_precommit', full_name='RpbBucketProps.has_precommit', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='postcommit', full_name='RpbBucketProps.postcommit', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='has_postcommit', full_name='RpbBucketProps.has_postcommit', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chash_keyfun', full_name='RpbBucketProps.chash_keyfun', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='linkfun', full_name='RpbBucketProps.linkfun', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='old_vclock', full_name='RpbBucketProps.old_vclock', index=9,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='young_vclock', full_name='RpbBucketProps.young_vclock', index=10,
      number=11, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='big_vclock', full_name='RpbBucketProps.big_vclock', index=11,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='small_vclock', full_name='RpbBucketProps.small_vclock', index=12,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pr', full_name='RpbBucketProps.pr', index=13,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='r', full_name='RpbBucketProps.r', index=14,
      number=15, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='w', full_name='RpbBucketProps.w', index=15,
      number=16, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pw', full_name='RpbBucketProps.pw', index=16,
      number=17, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dw', full_name='RpbBucketProps.dw', index=17,
      number=18, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rw', full_name='RpbBucketProps.rw', index=18,
      number=19, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='basic_quorum', full_name='RpbBucketProps.basic_quorum', index=19,
      number=20, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='notfound_ok', full_name='RpbBucketProps.notfound_ok', index=20,
      number=21, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='backend', full_name='RpbBucketProps.backend', index=21,
      number=22, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='search', full_name='RpbBucketProps.search', index=22,
      number=23, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='repl', full_name='RpbBucketProps.repl', index=23,
      number=24, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='search_index', full_name='RpbBucketProps.search_index', index=24,
      number=25, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='datatype', full_name='RpbBucketProps.datatype', index=25,
      number=26, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='consistent', full_name='RpbBucketProps.consistent', index=26,
      number=27, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='write_once', full_name='RpbBucketProps.write_once', index=27,
      number=28, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RPBBUCKETPROPS_RPBREPLMODE,
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=610,
  serialized_end=1298,
)


_RPBAUTHREQ = _descriptor.Descriptor(
  name='RpbAuthReq',
  full_name='RpbAuthReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='RpbAuthReq.user', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='password', full_name='RpbAuthReq.password', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1300,
  serialized_end=1344,
)


_RPBTOGGLEENCODINGREQ = _descriptor.Descriptor(
  name='RpbToggleEncodingReq',
  full_name='RpbToggleEncodingReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='use_native', full_name='RpbToggleEncodingReq.use_native', index=0,
      number=1, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1346,
  serialized_end=1388,
)


_RPBTOGGLEENCODINGRESP = _descriptor.Descriptor(
  name='RpbToggleEncodingResp',
  full_name='RpbToggleEncodingResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='use_native', full_name='RpbToggleEncodingResp.use_native', index=0,
      number=1, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1390,
  serialized_end=1433,
)

_RPBGETBUCKETRESP.fields_by_name['props'].message_type = _RPBBUCKETPROPS
_RPBSETBUCKETREQ.fields_by_name['props'].message_type = _RPBBUCKETPROPS
_RPBSETBUCKETTYPEREQ.fields_by_name['props'].message_type = _RPBBUCKETPROPS
_RPBCOMMITHOOK.fields_by_name['modfun'].message_type = _RPBMODFUN
_RPBBUCKETPROPS.fields_by_name['precommit'].message_type = _RPBCOMMITHOOK
_RPBBUCKETPROPS.fields_by_name['postcommit'].message_type = _RPBCOMMITHOOK
_RPBBUCKETPROPS.fields_by_name['chash_keyfun'].message_type = _RPBMODFUN
_RPBBUCKETPROPS.fields_by_name['linkfun'].message_type = _RPBMODFUN
_RPBBUCKETPROPS.fields_by_name['repl'].enum_type = _RPBBUCKETPROPS_RPBREPLMODE
_RPBBUCKETPROPS_RPBREPLMODE.containing_type = _RPBBUCKETPROPS;
DESCRIPTOR.message_types_by_name['RpbErrorResp'] = _RPBERRORRESP
DESCRIPTOR.message_types_by_name['RpbGetServerInfoResp'] = _RPBGETSERVERINFORESP
DESCRIPTOR.message_types_by_name['RpbPair'] = _RPBPAIR
DESCRIPTOR.message_types_by_name['RpbGetBucketReq'] = _RPBGETBUCKETREQ
DESCRIPTOR.message_types_by_name['RpbGetBucketResp'] = _RPBGETBUCKETRESP
DESCRIPTOR.message_types_by_name['RpbSetBucketReq'] = _RPBSETBUCKETREQ
DESCRIPTOR.message_types_by_name['RpbResetBucketReq'] = _RPBRESETBUCKETREQ
DESCRIPTOR.message_types_by_name['RpbGetBucketTypeReq'] = _RPBGETBUCKETTYPEREQ
DESCRIPTOR.message_types_by_name['RpbSetBucketTypeReq'] = _RPBSETBUCKETTYPEREQ
DESCRIPTOR.message_types_by_name['RpbModFun'] = _RPBMODFUN
DESCRIPTOR.message_types_by_name['RpbCommitHook'] = _RPBCOMMITHOOK
DESCRIPTOR.message_types_by_name['RpbBucketProps'] = _RPBBUCKETPROPS
DESCRIPTOR.message_types_by_name['RpbAuthReq'] = _RPBAUTHREQ
DESCRIPTOR.message_types_by_name['RpbToggleEncodingReq'] = _RPBTOGGLEENCODINGREQ
DESCRIPTOR.message_types_by_name['RpbToggleEncodingResp'] = _RPBTOGGLEENCODINGRESP

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbErrorResp(_message.Message):
  DESCRIPTOR = _RPBERRORRESP

  # @@protoc_insertion_point(class_scope:RpbErrorResp)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbGetServerInfoResp(_message.Message):
  DESCRIPTOR = _RPBGETSERVERINFORESP

  # @@protoc_insertion_point(class_scope:RpbGetServerInfoResp)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbPair(_message.Message):
  DESCRIPTOR = _RPBPAIR

  # @@protoc_insertion_point(class_scope:RpbPair)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbGetBucketReq(_message.Message):
  DESCRIPTOR = _RPBGETBUCKETREQ

  # @@protoc_insertion_point(class_scope:RpbGetBucketReq)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbGetBucketResp(_message.Message):
  DESCRIPTOR = _RPBGETBUCKETRESP

  # @@protoc_insertion_point(class_scope:RpbGetBucketResp)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbSetBucketReq(_message.Message):
  DESCRIPTOR = _RPBSETBUCKETREQ

  # @@protoc_insertion_point(class_scope:RpbSetBucketReq)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbResetBucketReq(_message.Message):
  DESCRIPTOR = _RPBRESETBUCKETREQ

  # @@protoc_insertion_point(class_scope:RpbResetBucketReq)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbGetBucketTypeReq(_message.Message):
  DESCRIPTOR = _RPBGETBUCKETTYPEREQ

  # @@protoc_insertion_point(class_scope:RpbGetBucketTypeReq)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbSetBucketTypeReq(_message.Message):
  DESCRIPTOR = _RPBSETBUCKETTYPEREQ

  # @@protoc_insertion_point(class_scope:RpbSetBucketTypeReq)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbModFun(_message.Message):
  DESCRIPTOR = _RPBMODFUN

  # @@protoc_insertion_point(class_scope:RpbModFun)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbCommitHook(_message.Message):
  DESCRIPTOR = _RPBCOMMITHOOK

  # @@protoc_insertion_point(class_scope:RpbCommitHook)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbBucketProps(_message.Message):
  DESCRIPTOR = _RPBBUCKETPROPS

  # @@protoc_insertion_point(class_scope:RpbBucketProps)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbAuthReq(_message.Message):
  DESCRIPTOR = _RPBAUTHREQ

  # @@protoc_insertion_point(class_scope:RpbAuthReq)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbToggleEncodingReq(_message.Message):
  DESCRIPTOR = _RPBTOGGLEENCODINGREQ

  # @@protoc_insertion_point(class_scope:RpbToggleEncodingReq)

@add_metaclass(_reflection.GeneratedProtocolMessageType)
class RpbToggleEncodingResp(_message.Message):
  DESCRIPTOR = _RPBTOGGLEENCODINGRESP

  # @@protoc_insertion_point(class_scope:RpbToggleEncodingResp)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), '\n\027com.basho.riak.protobufB\006RiakPB')
# @@protoc_insertion_point(module_scope)



"""
Copyright 2013 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from collections import namedtuple, Sequence


CONTINUATION = namedtuple('Continuation', ['c'])


class IndexPage(Sequence, object):
    """
    Encapsulates a single page of results from a secondary index
    query, with the ability to iterate over results (if not streamed),
    capture the page marker (continuation), and automatically fetch
    the next page.

    While users will interact with this object, it will be created
    automatically by the client and does not need to be instantiated
    elsewhere.
    """
    def __init__(self, client, bucket, index, startkey, endkey, return_terms,
                 max_results, term_regex):
        self.client = client
        self.bucket = bucket
        self.index = index
        self.startkey = startkey
        self.endkey = endkey
        self.return_terms = return_terms
        self.max_results = max_results
        self.results = None
        self.stream = False
        self.term_regex = term_regex

    continuation = None
    """
    The opaque page marker that is used when fetching the next chunk
    of results. The user can simply call :meth:`next_page` to do so,
    or pass this to the :meth:`~riak.client.RiakClient.get_index`
    method using the ``continuation`` option.
    """

    def __iter__(self):
        """
        Emulates the iterator interface. When streaming, this means
        delegating to the stream, otherwise iterating over the
        existing result set.
        """
        if self.results is None:
            raise ValueError("No index results to iterate")

        try:
            for result in self.results:
                if self.stream and isinstance(result, CONTINUATION):
                    self.continuation = result.c
                else:
                    yield self._inject_term(result)
        finally:
            if self.stream:
                self.results.close()

    def __len__(self):
        """
        Returns the length of the captured results.
        """
        if self._has_results():
            return len(self.results)
        else:
            raise ValueError("Streamed index page has no length")

    def __getitem__(self, index):
        """
        Fetches an item by index from the captured results.
        """
        if self._has_results():
            return self.results[index]
        else:
            raise ValueError("Streamed index page has no entries")

    def __eq__(self, other):
        """
        An IndexPage can pretend to be equal to a list when it has
        captured results by simply comparing the internal results to
        the passed list. Otherwise the other object needs to be an
        equivalent IndexPage.
        """
        if isinstance(other, list) and self._has_results():
            return self._inject_term(self.results) == other
        elif isinstance(other, IndexPage):
            return other.__dict__ == self.__dict__
        else:
            return False

    def __ne__(self, other):
        """
        Converse of __eq__.
        """
        return not self.__eq__(other)

    def has_next_page(self):
        """
        Whether there is another page available, i.e. the response
        included a continuation.
        """
        return self.continuation is not None

    def next_page(self, timeout=None, stream=None):
        """
        Fetches the next page using the same parameters as the
        original query.

        Note that if streaming was used before, it will be used again
        unless overridden.

        :param stream: whether to enable streaming. `True` enables,
            `False` disables, `None` uses previous value.
        :type stream: boolean
        :param timeout: a timeout value in milliseconds, or 'infinity'
        :type timeout: int
        """
        if not self.continuation:
            raise ValueError("Cannot get next index page, no continuation")

        if stream is not None:
            self.stream = stream

        args = {'bucket': self.bucket,
                'index': self.index,
                'startkey': self.startkey,
                'endkey': self.endkey,
                'return_terms': self.return_terms,
                'max_results': self.max_results,
                'continuation': self.continuation,
                'timeout': timeout,
                'term_regex': self.term_regex}

        if self.stream:
            return self.client.stream_index(**args)
        else:
            return self.client.get_index(**args)

    def _has_results(self):
        """
        When not streaming, have results been assigned?
        """
        return not (self.stream or self.results is None)

    def _should_inject_term(self, term):
        """
        The index term should be injected when using an equality query
        and the return terms option. If the term is already a tuple,
        it can be skipped.
        """
        return self.return_terms and not self.endkey

    def _inject_term(self, result):
        """
        Upgrades a result (streamed or not) to include the index term
        when an equality query is used with return_terms.
        """
        if self._should_inject_term(result):
            if type(result) is list:
                return [(self.startkey, r) for r in result]
            else:
                return (self.startkey, result)
        else:
            return result

    def __repr__(self):
        return "<{!s} {!r}>".format(self.__class__.__name__, self.__dict__)

    def close(self):
        if self.stream:
            self.results.close()



# -*- coding: utf-8 -*-
from __future__ import print_function

import platform

from six import PY2
from riak.mapreduce import RiakMapReduce
from riak import key_filter, RiakError
from riak.tests import RUN_YZ
from riak.tests.base import IntegrationTestBase
from riak.tests.test_yokozuna import wait_for_yz_index
from riak.tests import RUN_SECURITY
from riak.tests.yz_setup import yzSetUp, yzTearDown

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest


testrun_yz_mr = {'btype': 'mr',
                 'bucket': 'mrbucket',
                 'index': 'mrbucket'}


def setUpModule():
    yzSetUp(testrun_yz_mr)


def tearDownModule():
    yzTearDown(testrun_yz_mr)


class LinkTests(IntegrationTestBase, unittest.TestCase):
    def test_store_and_get_links(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        if PY2:
            bucket.new(key=self.key_name, encoded_data='2',
                       content_type='application/octet-stream') \
                .add_link(bucket.new("foo1")) \
                .add_link(bucket.new("foo2"), "tag") \
                .add_link(bucket.new("foo3"), "tag2!@#%^&*)") \
                .store()
        else:
            bucket.new(key=self.key_name, data='2',
                       content_type='application/octet-stream') \
                .add_link(bucket.new("foo1")) \
                .add_link(bucket.new("foo2"), "tag") \
                .add_link(bucket.new("foo3"), "tag2!@#%^&*)") \
                .store()
        obj = bucket.get(self.key_name)
        links = obj.links
        self.assertEqual(len(links), 3)
        for bucket, key, tag in links:
            if (key == "foo1"):
                self.assertEqual(bucket, self.bucket_name)
            elif (key == "foo2"):
                self.assertEqual(tag, "tag")
            elif (key == "foo3"):
                self.assertEqual(tag, "tag2!@#%^&*)")
            else:
                self.assertEqual(key, "unknown key")

    def test_set_links(self):
        # Create the object
        bucket = self.client.bucket(self.bucket_name)
        o = bucket.new(self.key_name, 2)
        o.links = [(self.bucket_name, "foo1", None),
                   (self.bucket_name, "foo2", "tag"),
                   ("bucket", "foo2", "tag2")]
        o.store()
        obj = bucket.get(self.key_name)
        links = sorted(obj.links, key=lambda x: x[1])
        self.assertEqual(len(links), 3)
        self.assertEqual(links[0][1], "foo1")
        self.assertEqual(links[1][1], "foo2")
        self.assertEqual(links[1][2], "tag")
        self.assertEqual(links[2][1], "foo2")
        self.assertEqual(links[2][2], "tag2")

    # "Link walking is deprecated in Riak 2.0 and is not compatible
    #  with security."
    @unittest.skipIf(RUN_SECURITY, 'RUN_SECURITY is set')
    def test_link_walking(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2) \
            .add_link(bucket.new("foo1", "test1").store()) \
            .add_link(bucket.new("foo2", "test2").store(), "tag") \
            .add_link(bucket.new("foo3", "test3").store(), "tag2!@#%^&*)") \
            .store()
        obj = bucket.get("foo")
        results = obj.link(self.bucket_name).run()
        self.assertEqual(len(results), 3)
        results = obj.link(self.bucket_name, "tag").run()
        self.assertEqual(len(results), 1)


class ErlangMapReduceTests(IntegrationTestBase, unittest.TestCase):
    def test_erlang_map_reduce(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        bucket.new("bar", 2).store()
        bucket.new("baz", 4).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo") \
            .add(self.bucket_name, "bar") \
            .add(self.bucket_name, "baz") \
            .map(["riak_kv_mapreduce", "map_object_value"]) \
            .reduce(["riak_kv_mapreduce", "reduce_set_union"]) \
            .run()
        self.assertEqual(len(result), 2)

    def test_erlang_map_reduce_bucket_type(self):
        # Create the object...
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        bucket.new("bar", 2).store()
        bucket.new("baz", 4).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo", bucket_type='no_siblings') \
            .add(self.bucket_name, "bar", bucket_type='no_siblings') \
            .add(self.bucket_name, "baz", bucket_type='no_siblings') \
            .map(["riak_kv_mapreduce", "map_object_value"]) \
            .reduce(["riak_kv_mapreduce", "reduce_set_union"]) \
            .run()
        self.assertEqual(len(result), 2)

    def test_erlang_source_map_reduce(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        bucket.new("bar", 3).store()
        bucket.new("baz", 4).store()
        strfun_allowed = True
        result = []
        # Run the map...
        try:
            result = self.client \
                .add(self.bucket_name, "foo") \
                .add(self.bucket_name, "bar") \
                .add(self.bucket_name, "baz") \
                .map("""fun(Object, _KD, _A) ->
            Value = riak_object:get_value(Object),
            [Value]
        end.""", {'language': 'erlang'}).run()
        except RiakError as e:
            if e.value.startswith('May have tried'):
                strfun_allowed = False
            else:
                print("test_erlang_source_map_reduce {}".format(e.value))
        if strfun_allowed:
            self.assertEqual(result, ['2', '3', '4'])

    def test_erlang_source_map_reduce_bucket_type(self):
        # Create the object...
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        bucket.new("bar", 3).store()
        bucket.new("baz", 4).store()
        strfun_allowed = True
        # Run the map...
        try:
            result = self.client \
                .add(self.bucket_name, "foo", bucket_type='no_siblings') \
                .add(self.bucket_name, "bar", bucket_type='no_siblings') \
                .add(self.bucket_name, "baz", bucket_type='no_siblings') \
                .map("""fun(Object, _KD, _A) ->
            Value = riak_object:get_value(Object),
            [Value]
        end.""", {'language': 'erlang'}).run()
        except RiakError as e:
            if e.value.startswith('May have tried'):
                strfun_allowed = False
        if strfun_allowed:
            self.assertEqual(result, ['2', '3', '4'])

    def test_client_exceptional_paths(self):
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        bucket.new("bar", 2).store()
        bucket.new("baz", 4).store()

        # adding a b-key pair to a bucket input
        with self.assertRaises(ValueError):
            mr = self.client.add(self.bucket_name)
            mr.add(self.bucket_name, 'bar')

        # adding a b-key pair to a query input
        with self.assertRaises(ValueError):
            mr = self.client.search(self.bucket_name, 'fleh')
            mr.add(self.bucket_name, 'bar')

        # adding a key filter to a query input
        with self.assertRaises(ValueError):
            mr = self.client.search(self.bucket_name, 'fleh')
            mr.add_key_filter("tokenize", "-", 1)


class JSMapReduceTests(IntegrationTestBase, unittest.TestCase):

    def test_javascript_source_map(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        # Run the map...
        mr = self.client.add(self.bucket_name, "foo")
        result = mr.map(
            "function (v) { return [JSON.parse(v.values[0].data)]; }").run()
        self.assertEqual(result, [2])

        # test ASCII-encodable unicode is accepted
        mr.map(u"function (v) { return [JSON.parse(v.values[0].data)]; }")

        # test non-ASCII-encodable unicode is rejected in Python 2.x
        if PY2:
            self.assertRaises(TypeError, mr.map,
                              u"""
                              function (v) {
                              /* æ */
                                return [JSON.parse(v.values[0].data)];
                              }""")
        else:
            mr = self.client.add(self.bucket_name, "foo")
            result = mr.map("""function (v) {
                      /* æ */
                        return [JSON.parse(v.values[0].data)];
                      }""").run()
            self.assertEqual(result, [2])

        # test non-ASCII-encodable string is rejected in Python 2.x
        if PY2:
            self.assertRaises(TypeError, mr.map,
                              """function (v) {
                                   /* æ */
                                   return [JSON.parse(v.values[0].data)];
                                 }""")

    def test_javascript_named_map(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo") \
            .map("Riak.mapValuesJson") \
            .run()
        self.assertEqual(result, [2])

    def test_javascript_named_map_bucket_type(self):
        # Create the object...
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo", bucket_type='no_siblings') \
            .map("Riak.mapValuesJson") \
            .run()
        self.assertEqual(result, [2])

    def test_javascript_source_map_reduce(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        bucket.new("bar", 3).store()
        bucket.new("baz", 4).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo") \
            .add(self.bucket_name, "bar") \
            .add(self.bucket_name, "baz") \
            .map("function (v) { return [1]; }") \
            .reduce("Riak.reduceSum") \
            .run()
        self.assertEqual(result, [3])

    def test_javascript_source_map_reduce_bucket_type(self):
        # Create the object...
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        bucket.new("bar", 3).store()
        bucket.new("baz", 4).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo", bucket_type='no_siblings') \
            .add(self.bucket_name, "bar", bucket_type='no_siblings') \
            .add(self.bucket_name, "baz", bucket_type='no_siblings') \
            .map("function (v) { return [1]; }") \
            .reduce("Riak.reduceSum") \
            .run()
        self.assertEqual(result, [3])

    def test_javascript_named_map_reduce(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        bucket.new("bar", 3).store()
        bucket.new("baz", 4).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo") \
            .add(self.bucket_name, "bar") \
            .add(self.bucket_name, "baz") \
            .map("Riak.mapValuesJson") \
            .reduce("Riak.reduceSum") \
            .run()
        self.assertEqual(result, [9])

    def test_javascript_named_map_reduce_bucket_type(self):
        # Create the object...
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        bucket.new("bar", 3).store()
        bucket.new("baz", 4).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo", bucket_type='no_siblings') \
            .add(self.bucket_name, "bar", bucket_type='no_siblings') \
            .add(self.bucket_name, "baz", bucket_type='no_siblings') \
            .map("Riak.mapValuesJson") \
            .reduce("Riak.reduceSum") \
            .run()
        self.assertEqual(result, [9])

    def test_javascript_bucket_map_reduce(self):
        # Create the object...
        bucket = self.client.bucket("bucket_%s" % self.randint())
        bucket.new("foo", 2).store()
        bucket.new("bar", 3).store()
        bucket.new("baz", 4).store()
        # Run the map...
        result = self.client \
            .add(bucket.name) \
            .map("Riak.mapValuesJson") \
            .reduce("Riak.reduceSum") \
            .run()
        self.assertEqual(result, [9])

    def test_javascript_bucket_map_reduceP_bucket_type(self):
        # Create the object...
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket("bucket_%s" % self.randint())
        bucket.new("foo", 2).store()
        bucket.new("bar", 3).store()
        bucket.new("baz", 4).store()
        # Run the map...
        result = self.client \
            .add(bucket.name, bucket_type='no_siblings') \
            .map("Riak.mapValuesJson") \
            .reduce("Riak.reduceSum") \
            .run()
        self.assertEqual(result, [9])

    def test_javascript_arg_map_reduce(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo", 5) \
            .add(self.bucket_name, "foo", 10) \
            .add(self.bucket_name, "foo", 15) \
            .add(self.bucket_name, "foo", -15) \
            .add(self.bucket_name, "foo", -5) \
            .map("function(v, arg) { return [arg]; }") \
            .reduce("Riak.reduceSum") \
            .run()
        self.assertEqual(result, [10])

    def test_javascript_arg_map_reduce_bucket_type(self):
        # Create the object...
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        # Run the map...
        result = self.client \
            .add(self.bucket_name, "foo", 5, bucket_type='no_siblings') \
            .add(self.bucket_name, "foo", 10, bucket_type='no_siblings') \
            .add(self.bucket_name, "foo", 15, bucket_type='no_siblings') \
            .add(self.bucket_name, "foo", -15, bucket_type='no_siblings') \
            .add(self.bucket_name, "foo", -5, bucket_type='no_siblings') \
            .map("function(v, arg) { return [arg]; }") \
            .reduce("Riak.reduceSum") \
            .run()
        self.assertEqual(result, [10])

    def test_key_filters(self):
        bucket = self.client.bucket("kftest")
        bucket.new("basho-20101215", 1).store()
        bucket.new("google-20110103", 2).store()
        bucket.new("yahoo-20090613", 3).store()

        result = self.client \
            .add("kftest") \
            .add_key_filters([["tokenize", "-", 2]]) \
            .add_key_filter("ends_with", "0613") \
            .map("function (v, keydata) { return [v.key]; }") \
            .run()

        self.assertEqual(result, ["yahoo-20090613"])

    def test_key_filters_bucket_type(self):
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket("kftest")
        bucket.new("basho-20101215", 1).store()
        bucket.new("google-20110103", 2).store()
        bucket.new("yahoo-20090613", 3).store()

        result = self.client \
            .add("kftest", bucket_type='no_siblings') \
            .add_key_filters([["tokenize", "-", 2]]) \
            .add_key_filter("ends_with", "0613") \
            .map("function (v, keydata) { return [v.key]; }") \
            .run()

        self.assertEqual(result, ["yahoo-20090613"])

    def test_key_filters_f_chain(self):
        bucket = self.client.bucket("kftest")
        bucket.new("basho-20101215", 1).store()
        bucket.new("google-20110103", 2).store()
        bucket.new("yahoo-20090613", 3).store()

        # compose a chain of key filters using f as the root of
        # two filters ANDed together to ensure that f can be the root
        # of multiple chains
        filters = key_filter.tokenize("-", 1).eq("yahoo") \
            & key_filter.tokenize("-", 2).ends_with("0613")

        result = self.client \
            .add("kftest") \
            .add_key_filters(filters) \
            .map("function (v, keydata) { return [v.key]; }") \
            .run()

        self.assertEqual(result, ["yahoo-20090613"])

    def test_key_filters_with_search_query(self):
        mapreduce = self.client.search("kftest", "query")
        self.assertRaises(Exception, mapreduce.add_key_filters,
                          [["tokenize", "-", 2]])
        self.assertRaises(Exception, mapreduce.add_key_filter,
                          "ends_with", "0613")

    def test_map_reduce_from_object(self):
        # Create the object...
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("foo", 2).store()
        obj = bucket.get("foo")
        result = obj.map("Riak.mapValuesJson").run()
        self.assertEqual(result, [2])

    def test_mr_list_add(self):
        bucket = self.client.bucket(self.bucket_name)
        for x in range(20):
            bucket.new('baz' + str(x),
                       'bazval' + str(x)).store()
        mr = self.client.add(self.bucket_name, ['baz' + str(x)
                                                for x in range(2, 5)])
        results = mr.map_values().run()
        results.sort()
        self.assertEqual(results,
                         [u'"bazval2"',
                          u'"bazval3"',
                          u'"bazval4"'])

    def test_mr_list_add_two_buckets(self):
        bucket = self.client.bucket(self.bucket_name)
        name2 = self.randname()
        for x in range(10):
            bucket.new('foo' + str(x),
                       'fooval' + str(x)).store()
        bucket = self.client.bucket(name2)
        for x in range(10):
            bucket.new('bar' + str(x),
                       'barval' + str(x)).store()

        mr = self.client.add(self.bucket_name, ['foo' + str(x)
                                                for x in range(2, 4)])
        mr.add(name2, ['bar' + str(x)
                       for x in range(5, 7)])
        results = mr.map_values().run()
        results.sort()

        self.assertEqual(results,
                         [u'"barval5"',
                          u'"barval6"',
                          u'"fooval2"',
                          u'"fooval3"'])

    def test_mr_list_add_mix(self):
        bucket = self.client.bucket("bucket_a")
        for x in range(10):
            bucket.new('foo' + str(x),
                       'fooval' + str(x)).store()
        bucket = self.client.bucket("bucket_b")
        for x in range(10):
            bucket.new('bar' + str(x),
                       'barval' + str(x)).store()

        mr = self.client.add('bucket_a', ['foo' + str(x)
                                          for x in range(2, 4)])
        mr.add('bucket_b', 'bar9')
        mr.add('bucket_b', 'bar2')
        results = mr.map_values().run()
        results.sort()

        self.assertEqual(results,
                         [u'"barval2"',
                          u'"barval9"',
                          u'"fooval2"',
                          u'"fooval3"'])

    @unittest.skipUnless(RUN_YZ, 'RUN_YZ is 0')
    def test_mr_search(self):
        """
        Try a successful map/reduce from search results.
        """
        btype = self.client.bucket_type(testrun_yz_mr['btype'])
        bucket = btype.bucket(testrun_yz_mr['bucket'])
        bucket.new("Pebbles", {"name_s": "Fruity Pebbles",
                               "maker_s": "Post",
                               "sugar_i": 9,
                               "calories_i": 110,
                               "fruit_b": True}).store()
        bucket.new("Loops", {"name_s": "Froot Loops",
                             "maker_s": "Kellogg's",
                             "sugar_i": 12,
                             "calories_i": 110,
                             "fruit_b": True}).store()
        bucket.new("Charms", {"name_s": "Lucky Charms",
                              "maker_s": "General Mills",
                              "sugar_i": 10,
                              "calories_i": 110,
                              "fruit_b": False}).store()
        bucket.new("Count", {"name_s": "Count Chocula",
                             "maker_s": "General Mills",
                             "sugar_i": 9,
                             "calories_i": 100,
                             "fruit_b": False}).store()
        bucket.new("Crunch", {"name_s": "Cap'n Crunch",
                              "maker_s": "Quaker Oats",
                              "sugar_i": 12,
                              "calories_i": 110,
                              "fruit_b": False}).store()
        # Wait for Solr to catch up
        wait_for_yz_index(bucket, "Crunch")
        mr = RiakMapReduce(self.client).search(testrun_yz_mr['bucket'],
                                               'fruit_b:false')
        mr.map("""function(v) {
            var solr_doc = JSON.parse(v.values[0].data);
            return [solr_doc["calories_i"]]; }""")
        result = mr.reduce('function(values, arg) ' +
                           '{ return [values.sort()[0]]; }').run()
        self.assertEqual(result, [100])


class MapReduceAliasTests(IntegrationTestBase, unittest.TestCase):
    """This tests the map reduce aliases"""

    def test_map_values(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        if PY2:
            bucket.new('one', encoded_data='value_1',
                       content_type='text/plain').store()
            bucket.new('two', encoded_data='value_2',
                       content_type='text/plain').store()
        else:
            bucket.new('one', data='value_1',
                       content_type='text/plain').store()
            bucket.new('two', data='value_2',
                       content_type='text/plain').store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values().run()

        # Sort the result so that we can have a consistent
        # expected value
        result.sort()

        self.assertEqual(result, ["value_1", "value_2"])

    def test_map_values_json(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data={'val': 'value_1'}).store()
        bucket.new('two', data={'val': 'value_2'}).store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values_json().run()

        # Sort the result so that we can have a consistent
        # expected value
        result.sort(key=lambda x: x['val'])

        self.assertEqual(result, [{'val': "value_1"}, {'val': "value_2"}])

    def test_reduce_sum(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data=1).store()
        bucket.new('two', data=2).store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values_json().reduce_sum().run()

        self.assertEqual(result, [3])

    def test_reduce_min(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data=1).store()
        bucket.new('two', data=2).store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values_json().reduce_min().run()

        self.assertEqual(result, [1])

    def test_reduce_max(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data=1).store()
        bucket.new('two', data=2).store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values_json().reduce_max().run()

        self.assertEqual(result, [2])

    def test_reduce_sort(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data="value1").store()
        bucket.new('two', data="value2").store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values_json().reduce_sort().run()

        self.assertEqual(result, ["value1", "value2"])

    def test_reduce_sort_custom(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data="value1").store()
        bucket.new('two', data="value2").store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values_json().reduce_sort("""function(x,y) {
           if(x == y) return 0;
           return x > y ? -1 : 1;
        }""").run()

        self.assertEqual(result, ["value2", "value1"])

    def test_reduce_numeric_sort(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data=1).store()
        bucket.new('two', data=2).store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values_json().reduce_numeric_sort().run()

        self.assertEqual(result, [1, 2])

    def test_reduce_limit(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data=1).store()
        bucket.new('two', data=2).store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values_json()\
                   .reduce_numeric_sort()\
                   .reduce_limit(1).run()

        self.assertEqual(result, [1])

    def test_reduce_slice(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data=1).store()
        bucket.new('two', data=2).store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')

        # Use the map_values alias
        result = mr.map_values_json()\
                   .reduce_numeric_sort()\
                   .reduce_slice(1, 2).run()

        self.assertEqual(result, [2])

    def test_filter_not_found(self):
        # Add a value to the bucket
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data=1).store()
        bucket.new('two', data=2).store()

        # Create a map reduce object and use one and two as inputs
        mr = self.client.add(self.bucket_name, 'one')\
                        .add(self.bucket_name, 'two')\
                        .add(self.bucket_name, self.key_name)

        # Use the map_values alias
        result = mr.map_values_json()\
                   .filter_not_found()\
                   .run()

        self.assertEqual(sorted(result), [1, 2])


class MapReduceStreamTests(IntegrationTestBase, unittest.TestCase):
    def test_stream_results(self):
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data=1).store()
        bucket.new('two', data=2).store()

        mr = RiakMapReduce(self.client).add(self.bucket_name, 'one')\
                                       .add(self.bucket_name, 'two')
        mr.map_values_json()
        results = []
        for phase, data in mr.stream():
            results.extend(data)

        self.assertEqual(sorted(results), [1, 2])

    def test_stream_cleanoperationsup(self):
        bucket = self.client.bucket(self.bucket_name)
        bucket.new('one', data=1).store()
        bucket.new('two', data=2).store()

        mr = RiakMapReduce(self.client).add(self.bucket_name, 'one')\
                                       .add(self.bucket_name, 'two')
        mr.map_values_json()
        try:
            for phase, data in mr.stream():
                raise RuntimeError("woops")
        except RuntimeError:
            pass

        # This should not raise an exception
        obj = bucket.get('one')
        if PY2:
            self.assertEqual('1', obj.encoded_data)
        else:
            self.assertEqual(b'1', obj.encoded_data)



"""
Copyright 2010 Rusty Klophaus <rusty@basho.com>
Copyright 2010 Justin Sheehy <justin@basho.com>
Copyright 2009 Jay Baird <jay@mochimedia.com>

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
from six import string_types, PY2
import mimetypes
from riak.util import lazy_property
from riak.datatypes import TYPES


def bucket_property(name, doc=None):
    def _prop_getter(self):
        return self.get_property(name)

    def _prop_setter(self, value):
        return self.set_property(name, value)

    return property(_prop_getter, _prop_setter, doc=doc)


class RiakBucket(object):
    """
    The ``RiakBucket`` object allows you to access and change information
    about a Riak bucket, and provides methods to create or retrieve
    objects within the bucket.
    """

    def __init__(self, client, name, bucket_type):
        """
        Returns a new ``RiakBucket`` instance.

        :param client: A :class:`RiakClient <riak.client.RiakClient>`
               instance
        :type client: :class:`RiakClient <riak.client.RiakClient>`
        :param name: The bucket name
        :type name: string
        :param bucket_type: The parent bucket type of this bucket
        :type bucket_type: :class:`BucketType`
        """

        if not isinstance(name, string_types):
            raise TypeError('Bucket name must be a string')

        if PY2:
            try:
                name = name.encode('ascii')
            except UnicodeError:
                raise TypeError('Unicode bucket names are not supported.')

        if not isinstance(bucket_type, BucketType):
            raise TypeError('Parent bucket type must be a BucketType instance')

        self._client = client
        self.name = name
        self.bucket_type = bucket_type
        self._encoders = {}
        self._decoders = {}
        self._resolver = None

    def __hash__(self):
        return hash((self.bucket_type.name, self.name, self._client))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return hash(self) == hash(other)
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return hash(self) != hash(other)
        else:
            return True

    def get_encoder(self, content_type):
        """
        Get the encoding function for the provided content type for
        this bucket.

        :param content_type: the requested media type
        :type content_type: str
        :param content_type: Content type requested
        """
        if content_type in self._encoders:
            return self._encoders[content_type]
        else:
            return self._client.get_encoder(content_type)

    def set_encoder(self, content_type, encoder):
        """
        Set the encoding function for the provided content type for
        this bucket.

        :param content_type: the requested media type
        :type content_type: str
        :param encoder: an encoding function, takes a single object
            argument and returns a string data as single argument.
        :type encoder: function
        """
        self._encoders[content_type] = encoder
        return self

    def get_decoder(self, content_type):
        """
        Get the decoding function for the provided content type for
        this bucket.

        :param content_type: the requested media type
        :type content_type: str
        :rtype: function
        """
        if content_type in self._decoders:
            return self._decoders[content_type]
        else:
            return self._client.get_decoder(content_type)

    def set_decoder(self, content_type, decoder):
        """
        Set the decoding function for the provided content type for
        this bucket.

        :param content_type: the requested media type
        :type content_type: str
        :param decoder: a decoding function, takes a string and
            returns a Python type
        :type decoder: function
        """
        self._decoders[content_type] = decoder
        return self

    def new(self, key=None, data=None, content_type='application/json',
            encoded_data=None):
        """A shortcut for manually instantiating a new
        :class:`~riak.riak_object.RiakObject` or a new
        :class:`~riak.datatypes.Datatype`, based on the presence and value
        of the :attr:`datatype <BucketType.datatype>` bucket property. When
        the bucket contains a :class:`~riak.datatypes.Datatype`, all
        arguments are ignored except ``key``, otherwise they are used to
        initialize the :class:`~riak.riak_object.RiakObject`.

        :param key: Name of the key. Leaving this to be None (default)
                    will make Riak generate the key on store.
        :type key: str
        :param data: The data to store in a
           :class:`~riak.riak_object.RiakObject`, see
           :attr:`RiakObject.data <riak.riak_object.RiakObject.data>`.
        :type data: object
        :param content_type: The media type of the data stored in the
           :class:`~riak.riak_object.RiakObject`, see
           :attr:`RiakObject.content_type
           <riak.riak_object.RiakObject.content_type>`.
        :type content_type: str
        :param encoded_data: The encoded data to store in a
           :class:`~riak.riak_object.RiakObject`, see
           :attr:`RiakObject.encoded_data
           <riak.riak_object.RiakObject.encoded_data>`.
        :type encoded_data: str
        :rtype: :class:`~riak.riak_object.RiakObject` or
                :class:`~riak.datatypes.Datatype`

        """
        from riak import RiakObject
        if self.bucket_type.datatype:
            return TYPES[self.bucket_type.datatype](bucket=self, key=key)

        if PY2:
            try:
                if isinstance(data, string_types):
                    data = data.encode('ascii')
            except UnicodeError:
                raise TypeError('Unicode data values are not supported.')

        obj = RiakObject(self._client, self, key)
        obj.content_type = content_type
        if data is not None:
            obj.data = data
        if encoded_data is not None:
            obj.encoded_data = encoded_data
        return obj

    def get(self, key, r=None, pr=None, timeout=None, include_context=None,
            basic_quorum=None, notfound_ok=None):
        """
        Retrieve a :class:`~riak.riak_object.RiakObject` or
        :class:`~riak.datatypes.Datatype`, based on the presence and value
        of the :attr:`datatype <BucketType.datatype>` bucket property.

        :param key: Name of the key.
        :type key: string
        :param r: R-Value of the request (defaults to bucket's R)
        :type r: integer
        :param pr: PR-Value of the request (defaults to bucket's PR)
        :type pr: integer
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :param include_context: if the bucket contains datatypes, include
           the opaque context in the result
        :type include_context: bool
        :param basic_quorum: whether to use the "basic quorum" policy
           for not-founds
        :type basic_quorum: bool
        :param notfound_ok: whether to treat not-found responses as successful
        :type notfound_ok: bool
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>` or
           :class:`~riak.datatypes.Datatype`

        """
        from riak import RiakObject
        if self.bucket_type.datatype:
            return self._client.fetch_datatype(self, key, r=r, pr=pr,
                                               timeout=timeout,
                                               include_context=include_context,
                                               basic_quorum=basic_quorum,
                                               notfound_ok=notfound_ok)
        else:
            obj = RiakObject(self._client, self, key)
            return obj.reload(r=r, pr=pr, timeout=timeout,
                              basic_quorum=basic_quorum,
                              notfound_ok=notfound_ok)

    def multiget(self, keys, r=None, pr=None, timeout=None,
                 basic_quorum=None, notfound_ok=None):
        """
        Retrieves a list of keys belonging to this bucket in parallel.

        :param keys: the keys to fetch
        :type keys: list
        :param r: R-Value for the requests (defaults to bucket's R)
        :type r: integer
        :param pr: PR-Value for the requests (defaults to bucket's PR)
        :type pr: integer
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :param basic_quorum: whether to use the "basic quorum" policy
           for not-founds
        :type basic_quorum: bool
        :param notfound_ok: whether to treat not-found responses as successful
        :type notfound_ok: bool
        :rtype: list of :class:`RiakObjects <riak.riak_object.RiakObject>`,
            :class:`Datatypes <riak.datatypes.Datatype>`, or tuples of
            bucket_type, bucket, key, and the exception raised on fetch
        """
        bkeys = [(self.bucket_type.name, self.name, key) for key in keys]
        return self._client.multiget(bkeys, r=r, pr=pr, timeout=timeout,
                                     basic_quorum=basic_quorum,
                                     notfound_ok=notfound_ok)

    def _get_resolver(self):
        if callable(self._resolver):
            return self._resolver
        elif self._resolver is None:
            return self._client.resolver
        else:
            raise TypeError("resolver is not a function")

    def _set_resolver(self, value):
        if value is None or callable(value):
            self._resolver = value
        else:
            raise TypeError("resolver is not a function")

    resolver = property(_get_resolver, _set_resolver,
                        doc="""The sibling-resolution function for this
                           bucket. If the resolver is not set, the
                           client's resolver will be used.""")

    n_val = bucket_property('n_val', doc="""
    N-value for this bucket, which is the number of replicas
    that will be written of each object in the bucket.

    .. warning:: Set this once before you write any data to the
        bucket, and never change it again, otherwise unpredictable
        things could happen. This should only be used if you know what
        you are doing.
    """)

    allow_mult = bucket_property('allow_mult', doc="""
    If set to True, then writes with conflicting data will be stored
    and returned to the client.

    :type bool: boolean
    """)

    r = bucket_property('r', doc="""
    The default 'read' quorum for this bucket (how many replicas must
    reply for a successful read). This should be an integer less than
    the 'n_val' property, or a string of 'one', 'quorum', 'all', or
    'default'""")

    pr = bucket_property('pr', doc="""
    The default 'primary read' quorum for this bucket (how many
    primary replicas are required for a successful read). This should
    be an integer less than the 'n_val' property, or a string of
    'one', 'quorum', 'all', or 'default'""")

    rw = bucket_property('rw', doc="""
    The default 'read' and 'write' quorum for this bucket (equivalent
    to 'r' and 'w' but for deletes). This should be an integer less
    than the 'n_val' property, or a string of 'one', 'quorum', 'all',
    or 'default'""")

    w = bucket_property('w', doc="""
    The default 'write' quorum for this bucket (how many replicas must
    acknowledge receipt of a write). This should be an integer less
    than the 'n_val' property, or a string of 'one', 'quorum', 'all',
    or 'default'""")

    dw = bucket_property('dw', doc="""
    The default 'durable write' quorum for this bucket (how many
    replicas must commit the write). This should be an integer less
    than the 'n_val' property, or a string of 'one', 'quorum', 'all',
    or 'default'""")

    pw = bucket_property('pw', doc="""
    The default 'primary write' quorum for this bucket (how many
    primary replicas are required for a successful write). This should
    be an integer less than the 'n_val' property, or a string of
    'one', 'quorum', 'all', or 'default'""")

    def set_property(self, key, value):
        """
        Set a bucket property.

        :param key: Property to set.
        :type key: string
        :param value: Property value.
        :type value: mixed
        """
        return self.set_properties({key: value})

    def get_property(self, key):
        """
        Retrieve a bucket property.

        :param key: The property to retrieve.
        :type key: string
        :rtype: mixed
        """
        return self.get_properties()[key]

    def set_properties(self, props):
        """
        Set multiple bucket properties in one call.

        :param props: A dictionary of properties
        :type props: dict
        """
        self._client.set_bucket_props(self, props)

    def get_properties(self):
        """
        Retrieve a dict of all bucket properties.

        :rtype: dict
        """
        return self._client.get_bucket_props(self)

    def clear_properties(self):
        """
        Reset all bucket properties to their defaults.
        """
        return self._client.clear_bucket_props(self)

    def get_keys(self):
        """
        Return all keys within the bucket.

        :rtype: list of keys
        """
        return self._client.get_keys(self)

    def stream_keys(self):
        """
        Streams all keys within the bucket through an iterator.

        The caller must close the stream when finished.  See
        :meth:`RiakClient.stream_keys()
        <riak.client.RiakClient.stream_keys>` for more details.

        :rtype: iterator
        """
        return self._client.stream_keys(self)

    def new_from_file(self, key, filename):
        """Create a new Riak object in the bucket, using the contents of
        the specified file. This is a shortcut for :meth:`new`, where the
        ``encoded_data`` and ``content_type`` are set for you.

        .. warning:: This is not supported for buckets that contain
           :class:`Datatypes <riak.datatypes.Datatype>`.

        :param key: the key of the new object
        :type key: string
        :param filename: the file to read the contents from
        :type filename: string
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        """
        binary_data = None
        with open(filename, 'rb') as f:
            binary_data = f.read()
        mimetype, encoding = mimetypes.guess_type(filename)
        if encoding:
            binary_data = bytearray(binary_data, encoding)
        else:
            binary_data = bytearray(binary_data)
        if not mimetype:
            mimetype = 'application/octet-stream'
        if PY2:
            return self.new(key, encoded_data=binary_data,
                            content_type=mimetype)
        else:
            return self.new(key, encoded_data=bytes(binary_data),
                            content_type=mimetype)

    def search_enabled(self):
        """
        Returns True if search indexing is enabled for this
        bucket.

        .. deprecated:: 2.1.0 (Riak 2.0)
           Use :ref:`Riak Search 2.0 <yz-label>` instead.
        """
        return self.get_properties().get('search', False)

    def enable_search(self):
        """
        Enable search indexing for this bucket.

        .. deprecated:: 2.1.0 (Riak 2.0)
           Use :ref:`Riak Search 2.0 <yz-label>` instead.
        """
        if not self.search_enabled():
            self.set_property('search', True)
        return True

    def disable_search(self):
        """
        Disable search indexing for this bucket.

        .. deprecated:: 2.1.0 (Riak 2.0)
           Use :ref:`Riak Search 2.0 <yz-label>` instead.
        """
        if self.search_enabled():
            self.set_property('search', False)
        return True

    def search(self, query, index=None, **params):
        """
        Queries a search index over objects in this bucket/index. See
        :meth:`RiakClient.fulltext_search()
        <riak.client.RiakClient.fulltext_search>` for more details.

        :param query: the search query
        :type query: string
        :param index: the index to search over. Defaults to the bucket's name.
        :type index: string or None
        :param params: additional query flags
        :type params: dict
        """
        search_index = index or self.name
        return self._client.fulltext_search(search_index, query, **params)

    def get_index(self, index, startkey, endkey=None, return_terms=None,
                  max_results=None, continuation=None, timeout=None,
                  term_regex=None):
        """
        Queries a secondary index over objects in this bucket,
        returning keys or index/key pairs. See
        :meth:`RiakClient.get_index()
        <riak.client.RiakClient.get_index>` for more details.
        """
        return self._client.get_index(self, index, startkey, endkey,
                                      return_terms=return_terms,
                                      max_results=max_results,
                                      continuation=continuation,
                                      timeout=timeout, term_regex=term_regex)

    def paginate_index(self, index, startkey, endkey=None,
                       return_terms=None, max_results=1000,
                       continuation=None, timeout=None, term_regex=None):
        """
        Paginates through a secondary index over objects in this bucket,
        returning keys or index/key pairs. See
        :meth:`RiakClient.paginate_index()
        <riak.client.RiakClient.paginate_index>` for more details.
        """
        return self._client.paginate_index(self, index, startkey, endkey,
                                           return_terms=return_terms,
                                           max_results=max_results,
                                           continuation=continuation,
                                           timeout=timeout,
                                           term_regex=term_regex)

    def stream_index(self, index, startkey, endkey=None, return_terms=None,
                     max_results=None, continuation=None, timeout=None,
                     term_regex=None):
        """
        Queries a secondary index over objects in this bucket,
        streaming keys or index/key pairs via an iterator.
        The caller must close the stream when finished.  See
        :meth:`RiakClient.stream_index()
        <riak.client.RiakClient.stream_index>` for more details.
        """
        return self._client.stream_index(self, index, startkey, endkey,
                                         return_terms=return_terms,
                                         max_results=max_results,
                                         continuation=continuation,
                                         timeout=timeout,
                                         term_regex=term_regex)

    def paginate_stream_index(self, index, startkey, endkey=None,
                              return_terms=None, max_results=1000,
                              continuation=None, timeout=None,
                              term_regex=None):
        """
        Paginates through a secondary index over objects in this bucket,
        streaming keys or index/key pairs. The caller must close the stream
        when finished.  See :meth:`RiakClient.paginate_stream_index()
        <riak.client.RiakClient.paginate_stream_index>` for more details.
        """
        return self._client.paginate_stream_index(self, index, startkey,
                                                  endkey,
                                                  return_terms=return_terms,
                                                  max_results=max_results,
                                                  continuation=continuation,
                                                  timeout=timeout,
                                                  term_regex=term_regex)

    def delete(self, key, **kwargs):
        """Deletes a key from Riak. Short hand for
        ``bucket.new(key).delete()``. See :meth:`RiakClient.delete()
        <riak.client.RiakClient.delete>` for options.

        :param key: The key for the object
        :type key: string
        :rtype: RiakObject
        """
        return self.new(key).delete(**kwargs)

    def get_counter(self, key, **kwargs):
        """
        Gets the value of a counter stored in this bucket. See
        :meth:`RiakClient.get_counter()
        <riak.client.RiakClient.get_counter>` for options.

        .. deprecated:: 2.1.0 (Riak 2.0) Riak 1.4-style counters are
           deprecated in favor of the :class:`~riak.datatypes.Counter`
           datatype.

        :param key: the key of the counter
        :type key: string
        :rtype: int
        """
        return self._client.get_counter(self, key, **kwargs)

    def update_counter(self, key, value, **kwargs):
        """
        Updates the value of a counter stored in this bucket. Positive
        values increment the counter, negative values decrement. See
        :meth:`RiakClient.update_counter()
        <riak.client.RiakClient.update_counter>` for options.

        .. deprecated:: 2.1.0 (Riak 2.0) Riak 1.4-style counters are
           deprecated in favor of the :class:`~riak.datatypes.Counter`
           datatype.

        :param key: the key of the counter
        :type key: string
        :param value: the amount to increment or decrement
        :type value: integer
        """
        return self._client.update_counter(self, key, value, **kwargs)

    increment_counter = update_counter

    def get_preflist(self, key):
        """
        Retrieve the preflist associated with a given bucket/key

        :param key: Name of the key.
        :type key: string
        :rtype: list of dict()
        """
        return self._client.get_preflist(self, key)

    def __str__(self):
        if self.bucket_type.is_default():
            return '<RiakBucket {0!r}>'.format(self.name)
        else:
            return '<RiakBucket {0!r}/{1!r}>'.format(self.bucket_type.name,
                                                     self.name)

    __repr__ = __str__


class BucketType(object):
    """
    The ``BucketType`` object allows you to access and change
    properties on a Riak bucket type and access buckets within its
    namespace.
    """
    def __init__(self, client, name):
        """
        Returns a new ``BucketType`` instance.

        :param client: A :class:`RiakClient <riak.client.RiakClient>`
               instance
        :type client: :class:`RiakClient <riak.client.RiakClient>`
        :param name: The bucket-type's name
        :type name: string
        """
        self._client = client
        self.name = name

    def is_default(self):
        """
        Whether this bucket type is the default type, or a user-defined type.

        :rtype: bool

        """
        return self.name == 'default'

    def get_property(self, key):
        """
        Retrieve a bucket-type property.

        :param key: The property to retrieve.
        :type key: string
        :rtype: mixed
        """
        return self.get_properties()[key]

    def set_property(self, key, value):
        """
        Set a bucket-type property.

        :param key: Property to set.
        :type key: string
        :param value: Property value.
        :type value: mixed
        """
        return self.set_properties({key: value})

    def get_properties(self):
        """
        Retrieve a dict of all bucket-type properties.

        :rtype: dict
        """
        return self._client.get_bucket_type_props(self)

    def set_properties(self, props):
        """
        Set multiple bucket-type properties in one call.

        :param props: A dictionary of properties
        :type props: dict
        """
        self._client.set_bucket_type_props(self, props)

    def bucket(self, name):
        """
        Gets a bucket that belongs to this bucket-type.

        :param name: the bucket name
        :type name: str
        :rtype: :class:`RiakBucket`
        """
        return self._client.bucket(name, self)

    def get_buckets(self, timeout=None):
        """
        Get the list of buckets under this bucket-type as
        :class:`RiakBucket <riak.bucket.RiakBucket>` instances.

        .. warning:: Do not use this in production, as it requires
           traversing through all keys stored in a cluster.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :rtype: list of :class:`RiakBucket <riak.bucket.RiakBucket>`
                instances
        """
        return self._client.get_buckets(bucket_type=self, timeout=timeout)

    def stream_buckets(self, timeout=None):
        """
        Streams the list of buckets under this bucket-type. This is a
        generator method that should be iterated over.

        The caller must close the stream when finished.  See
        :meth:`RiakClient.stream_buckets()
        <riak.client.RiakClient.stream_buckets>` for more details.

        .. warning:: Do not use this in production, as it requires
           traversing through all keys stored in a cluster.

        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :rtype: iterator that yields lists of :class:`RiakBucket
             <riak.bucket.RiakBucket>` instances
        """
        return self._client.stream_buckets(bucket_type=self, timeout=timeout)

    @lazy_property
    def datatype(self):
        """
        The assigned datatype for this bucket type, if present.

        :rtype: None or string
        """
        if self.is_default():
            return None
        else:
            return self.get_properties().get('datatype')

    def __str__(self):
        return "<BucketType {0!r}>".format(self.name)

    __repr__ = __str__

    def __hash__(self):
        return hash((self.name, self._client))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return hash(self) == hash(other)
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return hash(self) != hash(other)
        else:
            return True



"""
Copyright 2015 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from six import PY2
import base64
from riak.util import str_to_bytes
if PY2:
    from httplib import NotConnected, HTTPConnection
else:
    from http.client import NotConnected, HTTPConnection


class RiakHttpConnection(object):
    """
    Connection and low-level request methods for RiakHttpTransport.
    """

    def _request(self, method, uri, headers={}, body='', stream=False):
        """
        Given a Method, URL, Headers, and Body, perform and HTTP
        request, and return a 3-tuple containing the response status,
        response headers (as httplib.HTTPMessage), and response body.
        """
        response = None
        headers.setdefault('Accept',
                           'multipart/mixed, application/json, */*;q=0.5')

        if self._client._credentials:
            self._security_auth_headers(self._client._credentials.username,
                                        self._client._credentials.password,
                                        headers)

        try:
            self._connection.request(method, uri, body, headers)
            response = self._connection.getresponse()

            if stream:
                # The caller is responsible for fully reading the
                # response and closing it when streaming.
                response_body = response
            else:
                response_body = response.read()
        finally:
            if response and not stream:
                response.close()

        return response.status, response.msg, response_body

    def _connect(self):
        """
        Use the appropriate connection class; optionally with security.
        """
        timeout = None
        if self._options is not None and 'timeout' in self._options:
            timeout = self._options['timeout']

        if self._client._credentials:
            self._connection = self._connection_class(
                host=self._node.host,
                port=self._node.http_port,
                credentials=self._client._credentials,
                timeout=timeout)
        else:
            self._connection = self._connection_class(
                    host=self._node.host,
                    port=self._node.http_port,
                    timeout=timeout)
        # Forces the population of stats and resources before any
        # other requests are made.
        self.server_version

    def close(self):
        """
        Closes the underlying HTTP connection.
        """
        try:
            self._connection.close()
        except NotConnected:
            pass

    # These are set by the RiakHttpTransport initializer
    _connection_class = HTTPConnection
    _node = None

    def _security_auth_headers(self, username, password, headers):
        """
        Add in the requisite HTTP Authentication Headers

        :param username: Riak Security Username
        :type str
        :param password: Riak Security Password
        :type str
        :param headers: Dictionary of headers
        :type dict
        """
        userColonPassword = username + ":" + password
        b64UserColonPassword = base64. \
            b64encode(str_to_bytes(userColonPassword)).decode("ascii")
        headers['Authorization'] = 'Basic %s' % b64UserColonPassword



"""
Copyright 2012 Basho Technologies, Inc.
Copyright 2010 Rusty Klophaus <rusty@basho.com>
Copyright 2010 Justin Sheehy <justin@basho.com>
Copyright 2009 Jay Baird <jay@mochimedia.com>

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import errno
import socket
from riak.transports.pool import Pool
from riak.transports.pbc.transport import RiakPbcTransport


class RiakPbcPool(Pool):
    """
    A resource pool of PBC transports.
    """
    def __init__(self, client, **options):
        super(RiakPbcPool, self).__init__()
        self._client = client
        self._options = options

    def create_resource(self):
        node = self._client._choose_node()
        return RiakPbcTransport(node=node,
                                client=self._client,
                                **self._options)

    def destroy_resource(self, pbc):
        pbc.close()

# These are a specific set of socket errors
# that could be raised on send/recv that indicate
# that the socket is closed or reset, and is not
# usable. On seeing any of these errors, the socket
# should be closed, and the connection re-established.
CONN_CLOSED_ERRORS = (
    errno.EHOSTUNREACH,
    errno.ECONNRESET,
    errno.ECONNREFUSED,
    errno.ECONNABORTED,
    errno.ETIMEDOUT,
    errno.EBADF,
    errno.EPIPE
)


def is_retryable(err):
    """
    Determines if the given exception is something that is
    network/socket-related and should thus cause the PBC connection to
    close and the operation retried on another node.

    :rtype: boolean
    """
    if isinstance(err, socket.error):
        code = err.args[0]
        return code in CONN_CLOSED_ERRORS
    else:
        return False



import socket
import struct
import riak.pb.riak_pb2
import riak.pb.messages

from riak.security import SecurityError, USE_STDLIB_SSL
from riak import RiakError
from riak.util import bytes_to_str, str_to_bytes

from six import PY2
if not USE_STDLIB_SSL:
    from OpenSSL.SSL import Connection
    from riak.transports.security import configure_pyopenssl_context
else:
    import ssl
    from riak.transports.security import configure_ssl_context


class RiakPbcConnection(object):
    """
    Connection-related methods for RiakPbcTransport.
    """

    def _encode_msg(self, msg_code, msg=None):
        if msg is None:
            return struct.pack("!iB", 1, msg_code)
        msgstr = msg.SerializeToString()
        slen = len(msgstr)
        hdr = struct.pack("!iB", 1 + slen, msg_code)
        return hdr + msgstr

    def _request(self, msg_code, msg=None, expect=None):
        self._send_msg(msg_code, msg)
        return self._recv_msg(expect)

    def _non_connect_request(self, msg_code, msg=None, expect=None):
        """
        Similar to self._request, but doesn't try to initiate a connection,
        thus preventing an infinite loop.
        """
        self._non_connect_send_msg(msg_code, msg)
        return self._recv_msg(expect)

    def _non_connect_send_msg(self, msg_code, msg):
        """
        Similar to self._send, but doesn't try to initiate a connection,
        thus preventing an infinite loop.
        """
        self._socket.sendall(self._encode_msg(msg_code, msg))

    def _send_msg(self, msg_code, msg):
        self._connect()
        self._non_connect_send_msg(msg_code, msg)

    def _init_security(self):
        """
        Initialize a secure connection to the server.
        """
        if not self._starttls():
            raise SecurityError("Could not start TLS connection")
        # _ssh_handshake() will throw an exception upon failure
        self._ssl_handshake()
        if not self._auth():
            raise SecurityError("Could not authorize connection")

    def _starttls(self):
        """
        Exchange a STARTTLS message with Riak to initiate secure communications
        return True is Riak responds with a STARTTLS response, False otherwise
        """
        msg_code, _ = self._non_connect_request(
            riak.pb.messages.MSG_CODE_START_TLS)
        if msg_code == riak.pb.messages.MSG_CODE_START_TLS:
            return True
        else:
            return False

    def _auth(self):
        """
        Perform an authorization request against Riak
        returns True upon success, False otherwise
        Note: Riak will sleep for a short period of time upon a failed
              auth request/response to prevent denial of service attacks
        """
        req = riak.pb.riak_pb2.RpbAuthReq()
        req.user = str_to_bytes(self._client._credentials.username)
        password = self._client._credentials.password
        if not password:
            password = ''
        req.password = str_to_bytes(password)
        msg_code, _ = self._non_connect_request(
            riak.pb.messages.MSG_CODE_AUTH_REQ,
            req,
            riak.pb.messages.MSG_CODE_AUTH_RESP)
        if msg_code == riak.pb.messages.MSG_CODE_AUTH_RESP:
            return True
        else:
            return False

    if not USE_STDLIB_SSL:
        def _ssl_handshake(self):
            """
            Perform an SSL handshake w/ the server.
            Precondition: a successful STARTTLS exchange has
                         taken place with Riak
            returns True upon success, otherwise an exception is raised
            """
            if self._client._credentials:
                try:
                    ssl_ctx = configure_pyopenssl_context(self.
                                                          _client._credentials)
                    # attempt to upgrade the socket to SSL
                    ssl_socket = Connection(ssl_ctx, self._socket)
                    ssl_socket.set_connect_state()
                    ssl_socket.do_handshake()
                    # ssl handshake successful
                    self._socket = ssl_socket

                    self._client._credentials._check_revoked_cert(ssl_socket)
                    return True
                except Exception as e:
                    # fail if *any* exceptions are thrown during SSL handshake
                    raise SecurityError(e)
    else:
        def _ssl_handshake(self):
            """
            Perform an SSL handshake w/ the server.
            Precondition: a successful STARTTLS exchange has
                         taken place with Riak
            returns True upon success, otherwise an exception is raised
            """
            credentials = self._client._credentials
            if credentials:
                try:
                    ssl_ctx = configure_ssl_context(credentials)
                    host = self._address[0]
                    ssl_socket = ssl.SSLSocket(sock=self._socket,
                                               keyfile=credentials.pkey_file,
                                               certfile=credentials.cert_file,
                                               cert_reqs=ssl.CERT_REQUIRED,
                                               ca_certs=credentials.
                                               cacert_file,
                                               ciphers=credentials.ciphers,
                                               server_hostname=host)
                    ssl_socket.context = ssl_ctx
                    # ssl handshake successful
                    ssl_socket.do_handshake()
                    self._socket = ssl_socket

                    return True
                except ssl.SSLError as e:
                    raise SecurityError(e)
                except Exception as e:
                    # fail if *any* exceptions are thrown during SSL handshake
                    raise SecurityError(e)

    def _recv_msg(self, expect=None):
        self._recv_pkt()
        msg_code, = struct.unpack("B", self._inbuf[:1])
        if msg_code is riak.pb.messages.MSG_CODE_ERROR_RESP:
            err = self._parse_msg(msg_code, self._inbuf[1:])
            if err is None:
                raise RiakError('no error provided!')
            else:
                raise RiakError(bytes_to_str(err.errmsg))
        elif msg_code in riak.pb.messages.MESSAGE_CLASSES:
            msg = self._parse_msg(msg_code, self._inbuf[1:])
        else:
            raise Exception("unknown msg code %s" % msg_code)

        if expect and msg_code != expect:
            raise RiakError("unexpected protocol buffer message code: %d, %r"
                            % (msg_code, msg))
        return msg_code, msg

    def _recv_pkt(self):
        nmsglen = self._socket.recv(4)
        while len(nmsglen) < 4:
            x = self._socket.recv(4 - len(nmsglen))
            if not x:
                break
            nmsglen += x
        if len(nmsglen) != 4:
            raise RiakError(
                "Socket returned short packet length %d - expected 4"
                % len(nmsglen))
        msglen, = struct.unpack('!i', nmsglen)
        self._inbuf_len = msglen
        if PY2:
            self._inbuf = ''
        else:
            self._inbuf = bytes()
        while len(self._inbuf) < msglen:
            want_len = min(8192, msglen - len(self._inbuf))
            recv_buf = self._socket.recv(want_len)
            if not recv_buf:
                break
            self._inbuf += recv_buf
        if len(self._inbuf) != self._inbuf_len:
            raise RiakError("Socket returned short packet %d - expected %d"
                            % (len(self._inbuf), self._inbuf_len))

    def _connect(self):
        if not self._socket:
            if self._timeout:
                self._socket = socket.create_connection(self._address,
                                                        self._timeout)
            else:
                self._socket = socket.create_connection(self._address)
            if self._client._credentials:
                self._init_security()

    def close(self):
        """
        Closes the underlying socket of the PB connection.
        """
        if self._socket:
            self._socket.close()
            del self._socket

    def _parse_msg(self, code, packet):
        try:
            pbclass = riak.pb.messages.MESSAGE_CLASSES[code]
        except KeyError:
            pbclass = None

        if pbclass is None:
            return None

        pbo = pbclass()
        pbo.ParseFromString(packet)
        return pbo

    # These are set in the RiakPbcTransport initializer
    _address = None
    _timeout = None



# -*- coding: utf-8 -*-
import platform
import sys

from riak.tests import RUN_SECURITY, SECURITY_USER, SECURITY_PASSWD, \
    SECURITY_CACERT, SECURITY_KEY, SECURITY_CERT, SECURITY_REVOKED, \
    SECURITY_CERT_USER, SECURITY_BAD_CERT, SECURITY_CIPHERS
from riak.security import SecurityCreds
from riak.tests.base import IntegrationTestBase

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest


class SecurityTests(IntegrationTestBase, unittest.TestCase):
    @unittest.skipIf(RUN_SECURITY, 'RUN_SECURITY is 1')
    def test_security_disabled(self):
        """
        Test valid security settings without security enabled
        """
        topts = {'timeout': 1}
        # NB: can't use SECURITY_CREDS here since they won't be set
        # if RUN_SECURITY is UN-set
        creds = SecurityCreds(username='foo', password='bar')
        client = self.create_client(credentials=creds,
                                    transport_options=topts)
        myBucket = client.bucket('test')
        val1 = "foobar"
        key1 = myBucket.new('x', data=val1)
        with self.assertRaises(Exception):
            key1.store()
        client.close()

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_basic_connection(self):
        myBucket = self.client.bucket('test')
        val1 = "foobar"
        key1 = myBucket.new('x', data=val1)
        key1.store()
        myBucket.get('x')

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_bad_user(self):
        creds = SecurityCreds(username='foo',
                              password=SECURITY_PASSWD,
                              cacert_file=SECURITY_CACERT,
                              ciphers=SECURITY_CIPHERS)
        client = self.create_client(credentials=creds)
        with self.assertRaises(Exception):
            client.get_buckets()
        client.close()

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_bad_password(self):
        creds = SecurityCreds(username=SECURITY_USER,
                              password='foo',
                              cacert_file=SECURITY_CACERT,
                              ciphers=SECURITY_CIPHERS)
        client = self.create_client(credentials=creds)
        with self.assertRaises(Exception):
            client.get_buckets()
        client.close()

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_invalid_cert(self):
        creds = SecurityCreds(username=SECURITY_USER,
                              password=SECURITY_PASSWD,
                              cacert_file='/tmp/foo',
                              ciphers=SECURITY_CIPHERS)
        client = self.create_client(credentials=creds)
        with self.assertRaises(Exception):
            client.get_buckets()
        client.close()

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_password_without_cacert(self):
        creds = SecurityCreds(username=SECURITY_USER,
                              password=SECURITY_PASSWD,
                              ciphers=SECURITY_CIPHERS)
        client = self.create_client(credentials=creds)
        with self.assertRaises(Exception):
            myBucket = client.bucket('test')
            val1 = "foobar"
            key1 = myBucket.new('x', data=val1)
            key1.store()
        client.close()

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_cert_authentication(self):
        creds = SecurityCreds(username=SECURITY_CERT_USER,
                              ciphers=SECURITY_CIPHERS,
                              cert_file=SECURITY_CERT,
                              pkey_file=SECURITY_KEY,
                              cacert_file=SECURITY_CACERT)
        client = self.create_client(credentials=creds)
        myBucket = client.bucket('test')
        val1 = "foobar2"
        key1 = myBucket.new('x', data=val1)
        # Certificate Authentication is currently only supported
        # by Protocol Buffers
        if self.protocol == 'pbc':
            key1.store()
            myBucket.get('x')
        else:
            with self.assertRaises(Exception):
                key1.store()
                myBucket.get('x')
        client.close()

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_revoked_cert(self):
        creds = SecurityCreds(username=SECURITY_USER,
                              password=SECURITY_PASSWD,
                              ciphers=SECURITY_CIPHERS,
                              cacert_file=SECURITY_CACERT,
                              crl_file=SECURITY_REVOKED)
        # Currently Python >= 2.7.9 and Python 3.x native CRL doesn't seem to
        # work as advertised
        if sys.version_info >= (2, 7, 9):
            return
        client = self.create_client(credentials=creds)
        with self.assertRaises(Exception):
            client.get_buckets()
        client.close()

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_bad_ca_cert(self):
        creds = SecurityCreds(username=SECURITY_USER, password=SECURITY_PASSWD,
                              ciphers=SECURITY_CIPHERS,
                              cacert_file=SECURITY_BAD_CERT)
        client = self.create_client(credentials=creds)
        with self.assertRaises(Exception):
            client.get_buckets()
        client.close()

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_ciphers(self):
        creds = SecurityCreds(username=SECURITY_USER, password=SECURITY_PASSWD,
                              ciphers=SECURITY_CIPHERS,
                              cacert_file=SECURITY_CACERT)
        client = self.create_client(credentials=creds)
        myBucket = client.bucket('test')
        val1 = "foobar"
        key1 = myBucket.new('x', data=val1)
        key1.store()
        myBucket.get('x')
        client.close()

    @unittest.skipUnless(RUN_SECURITY, 'RUN_SECURITY is 0')
    def test_security_bad_ciphers(self):
        creds = SecurityCreds(username=SECURITY_USER, password=SECURITY_PASSWD,
                              cacert_file=SECURITY_CACERT,
                              ciphers='ECDHE-RSA-AES256-GCM-SHA384')
        client = self.create_client(credentials=creds)
        with self.assertRaises(Exception):
            client.get_buckets()
        client.close()



#!/usr/bin/env python

import platform
from setuptools import setup, find_packages
from version import get_version
from commands import setup_timeseries, build_messages

install_requires = ['six >= 1.8.0']
requires = ['six(>=1.8.0)']
if platform.python_version() < '2.7.9':
    install_requires.append("pyOpenSSL >= 0.14")
    requires.append("pyOpenSSL(>=0.14)")

if platform.python_version() < '3.0':
    install_requires.append('protobuf >=2.4.1, <2.7.0')
    requires.append('protobuf(>=2.4.1, <2.7.0)')
else:
    install_requires.append('python3_protobuf >=2.4.1, <2.6.0')
    requires.append('python3_protobuf(>=2.4.1, <2.6.0)')


tests_require = []
if platform.python_version() < '2.7.0':
    tests_require.append("unittest2")

setup(
    name='riak',
    version=get_version(),
    packages=find_packages(),
    requires=requires,
    install_requires=install_requires,
    tests_require=tests_require,
    package_data={'riak': ['erl_src/*']},
    description='Python client for Riak',
    zip_safe=True,
    options={'easy_install': {'allow_hosts': 'pypi.python.org'}},
    include_package_data=True,
    license='Apache 2',
    platforms='Platform Independent',
    author='Basho Technologies',
    author_email='clients@basho.com',
    test_suite='riak.tests.suite',
    url='https://github.com/basho/riak-python-client',
    cmdclass={
        'build_messages': build_messages,
        'setup_timeseries': setup_timeseries
    },
    classifiers=['License :: OSI Approved :: Apache Software License',
                 'Intended Audience :: Developers',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Topic :: Database']
    )



import os
import socket

from riak.test_server import TestServer
from riak.security import SecurityCreds

USE_TEST_SERVER = int(os.environ.get('USE_TEST_SERVER', '0'))
if USE_TEST_SERVER:
    HTTP_PORT = 9000
    PB_PORT = 9002
    test_server = TestServer()
    test_server.cleanup()
    test_server.prepare()
    test_server.start()

try:
    __import__('riak.pb')
    HAVE_PROTO = True
except ImportError:
    HAVE_PROTO = False


def hostname_resolves(hostname):
    try:
        socket.gethostbyname(hostname)
        return 1
    except socket.error:
        return 0

HOST = os.environ.get('RIAK_TEST_HOST', '127.0.0.1')

PROTOCOL = os.environ.get('RIAK_TEST_PROTOCOL', 'pbc')

PB_HOST = os.environ.get('RIAK_TEST_PB_HOST', HOST)
PB_PORT = int(os.environ.get('RIAK_TEST_PB_PORT', '10017'))

HTTP_HOST = os.environ.get('RIAK_TEST_HTTP_HOST', HOST)
HTTP_PORT = int(os.environ.get('RIAK_TEST_HTTP_PORT', '10018'))

# these ports are used to simulate errors, there shouldn't
# be anything listening on either port.
DUMMY_HTTP_PORT = int(os.environ.get('DUMMY_HTTP_PORT', '1023'))
DUMMY_PB_PORT = int(os.environ.get('DUMMY_PB_PORT', '1022'))

RUN_SEARCH = int(os.environ.get('RUN_SEARCH', '0'))
RUN_YZ = int(os.environ.get('RUN_YZ', '0'))

RUN_INDEXES = int(os.environ.get('RUN_INDEXES', '0'))

RUN_TIMESERIES = int(os.environ.get('RUN_TIMESERIES', '0'))

RUN_POOL = int(os.environ.get('RUN_POOL', '0'))
RUN_RESOLVE = int(os.environ.get('RUN_RESOLVE', '1'))
RUN_BTYPES = int(os.environ.get('RUN_BTYPES', '1'))
RUN_DATATYPES = int(os.environ.get('RUN_DATATYPES', '1'))

RUN_SECURITY = int(os.environ.get('RUN_SECURITY', '0'))
if RUN_SECURITY:
    h = 'riak-test'
    if hostname_resolves(h):
        HOST = PB_HOST = HTTP_HOST = h
    else:
        raise AssertionError(
                'RUN_SECURITY requires that the host name' +
                ' "riak-test" resolves to the IP address of a Riak node' +
                ' with security enabled.')

SECURITY_USER = os.environ.get('RIAK_TEST_SECURITY_USER', 'riakpass')
SECURITY_PASSWD = os.environ.get('RIAK_TEST_SECURITY_PASSWD', 'Test1234')

SECURITY_CACERT = os.environ.get('RIAK_TEST_SECURITY_CACERT',
                                 'tools/test-ca/certs/cacert.pem')
SECURITY_REVOKED = os.environ.get('RIAK_TEST_SECURITY_REVOKED',
                                  'tools/test-ca/crl/crl.pem')
SECURITY_BAD_CERT = os.environ.get('RIAK_TEST_SECURITY_BAD_CERT',
                                   'tools/test-ca/certs/badcert.pem')
# Certificate-based Authentication only supported by PBC
SECURITY_KEY = os.environ.get(
        'RIAK_TEST_SECURITY_KEY',
        'tools/test-ca/private/riakuser-client-cert-key.pem')
SECURITY_CERT = os.environ.get('RIAK_TEST_SECURITY_CERT',
                               'tools/test-ca/certs/riakuser-client-cert.pem')
SECURITY_CERT_USER = os.environ.get('RIAK_TEST_SECURITY_CERT_USER',
                                    'riakuser')

SECURITY_CIPHERS = 'DHE-RSA-AES128-SHA256:DHE-RSA-AES128-SHA:' + \
        'DHE-RSA-AES256-SHA256:DHE-RSA-AES256-SHA:AES128-SHA256:' + \
        'AES128-SHA:AES256-SHA256:AES256-SHA:RC4-SHA'

SECURITY_CREDS = None
if RUN_SECURITY:
    SECURITY_CREDS = SecurityCreds(username=SECURITY_USER,
                                   password=SECURITY_PASSWD,
                                   cacert_file=SECURITY_CACERT,
                                   ciphers=SECURITY_CIPHERS)



import logging

from riak import RiakError
from riak.tests import RUN_YZ
from riak.tests.base import IntegrationTestBase


def yzSetUp(*yzdata):
    if RUN_YZ:
        c = IntegrationTestBase.create_client()
        for yz in yzdata:
            logging.debug("yzSetUp: %s", yz)
            c.create_search_index(yz['index'], timeout=30000)
            if yz['btype'] is not None:
                t = c.bucket_type(yz['btype'])
                b = t.bucket(yz['bucket'])
            else:
                b = c.bucket(yz['bucket'])
            # Keep trying to set search bucket property until it succeeds
            index_set = False
            while not index_set:
                try:
                    b.set_property('search_index', yz['index'])
                    index_set = True
                except RiakError:
                    pass
        c.close()


def yzTearDown(c, *yzdata):
    if RUN_YZ:
        c = IntegrationTestBase.create_client()
        for yz in yzdata:
            logging.debug("yzTearDown: %s", yz)
            if yz['btype'] is not None:
                t = c.bucket_type(yz['btype'])
                b = t.bucket(yz['bucket'])
            else:
                b = c.bucket(yz['bucket'])
            b.set_property('search_index', '_dont_index_')
            c.delete_search_index(yz['index'])
            for keys in b.stream_keys():
                for key in keys:
                    b.delete(key)
        c.close()



"""
Copyright 2012 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import json
import re
from cgi import parse_header
from email import message_from_string
from riak.util import decode_index_value
from riak.client.index_page import CONTINUATION
from riak import RiakError
from six import PY2


class RiakHttpStream(object):
    """
    Base class for HTTP streaming iterators.
    """

    BLOCK_SIZE = 2048

    def __init__(self, response):
        self.response = response
        self.buffer = ''
        self.response_done = False
        self.resource = None

    def __iter__(self):
        return self

    def _read(self):
        chunk = self.response.read(self.BLOCK_SIZE)
        if PY2:
            if chunk == '':
                self.response_done = True
            self.buffer += chunk
        else:
            if chunk == b'':
                self.response_done = True
            self.buffer += chunk.decode('utf-8')

    def __next__(self):
        raise NotImplementedError

    def next(self):
        raise NotImplementedError

    def attach(self, resource):
        self.resource = resource

    def close(self):
        self.resource.release()


class RiakHttpJsonStream(RiakHttpStream):
    _json_field = None

    def next(self):
        # Python 2.x Version
        while '}' not in self.buffer and not self.response_done:
            self._read()

        if '}' in self.buffer:
            idx = self.buffer.index('}') + 1
            chunk = self.buffer[:idx]
            self.buffer = self.buffer[idx:]
            jsdict = json.loads(chunk)
            if 'error' in jsdict:
                self.close()
                raise RiakError(jsdict['error'])
            field = jsdict[self._json_field]
            return field
        else:
            raise StopIteration

    def __next__(self):
        # Python 3.x Version
        return self.next()


class RiakHttpKeyStream(RiakHttpJsonStream):
    """
    Streaming iterator for list-keys over HTTP
    """
    _json_field = u'keys'


class RiakHttpBucketStream(RiakHttpJsonStream):
    """
    Streaming iterator for list-buckets over HTTP
    """
    _json_field = u'buckets'


class RiakHttpMultipartStream(RiakHttpStream):
    """
    Streaming iterator for multipart messages over HTTP
    """
    def __init__(self, response):
        super(RiakHttpMultipartStream, self).__init__(response)
        ctypehdr = response.getheader('content-type')
        _, params = parse_header(ctypehdr)
        self.boundary_re = re.compile('\r?\n--%s(?:--)?\r?\n' %
                                      re.escape(params['boundary']))
        self.next_boundary = None
        self.seen_first = False

    def next(self):
        # multipart/mixed starts with a boundary, then the first part.
        if not self.seen_first:
            self.read_until_boundary()
            self.advance_buffer()
            self.seen_first = True

        self.read_until_boundary()

        if self.next_boundary:
            part = self.advance_buffer()
            message = message_from_string(part)
            return message
        else:
            raise StopIteration

    def __next__(self):
        # Python 3.x Version
        return self.next()

    def try_match(self):
        self.next_boundary = self.boundary_re.search(self.buffer)
        return self.next_boundary

    def advance_buffer(self):
        part = self.buffer[:self.next_boundary.start()]
        self.buffer = self.buffer[self.next_boundary.end():]
        self.next_boundary = None
        return part

    def read_until_boundary(self):
        while not self.try_match() and not self.response_done:
            self._read()


class RiakHttpMapReduceStream(RiakHttpMultipartStream):
    """
    Streaming iterator for MapReduce over HTTP
    """

    def next(self):
        message = super(RiakHttpMapReduceStream, self).next()
        payload = json.loads(message.get_payload())
        return payload['phase'], payload['data']

    def __next__(self):
        # Python 3.x Version
        return self.next()


class RiakHttpIndexStream(RiakHttpMultipartStream):
    """
    Streaming iterator for secondary indexes over HTTP
    """

    def __init__(self, response, index, return_terms):
        super(RiakHttpIndexStream, self).__init__(response)
        self.index = index
        self.return_terms = return_terms

    def next(self):
        message = super(RiakHttpIndexStream, self).next()
        payload = json.loads(message.get_payload())
        if u'error' in payload:
            raise RiakError(payload[u'error'])
        elif u'keys' in payload:
            return payload[u'keys']
        elif u'results' in payload:
            structs = payload[u'results']
            # Format is {"results":[{"2ikey":"primarykey"}, ...]}
            return [self._decode_pair(list(d.items())[0]) for d in structs]
        elif u'continuation' in payload:
            return CONTINUATION(payload[u'continuation'])

    def __next__(self):
        # Python 3.x Version
        return self.next()

    def _decode_pair(self, pair):
        return (decode_index_value(self.index, pair[0]), pair[1])



"""
Copyright 2012 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from __future__ import print_function
from contextlib import contextmanager
import threading


# This file is a rough port of the Innertube Ruby library
class BadResource(Exception):
    """
    Users of a :class:`Pool` should raise this error when the pool
    resource currently in-use is bad and should be removed from the
    pool.
    """
    pass


class Resource(object):
    """
    A member of the :class:`Pool`, a container for the actual resource
    being pooled and a marker for whether the resource is currently
    claimed.
    """
    def __init__(self, obj, pool):
        """
        Creates a new Resource, wrapping the passed object as the
        pooled resource.

        :param obj: the resource to wrap
        :type obj: object
        """

        self.object = obj
        """The wrapped pool resource."""

        self.claimed = False
        """Whether the resource is currently in use."""

        self.pool = pool
        """The pool that this resource belongs to."""

    def release(self):
        """
        Releases this resource back to the pool it came from.
        """
        self.pool.release(self)


class Pool(object):
    """
    A thread-safe, reentrant resource pool, ported from the
    "Innertube" Ruby library. Pool should be subclassed to implement
    the create_resource and destroy_resource functions that are
    responsible for creating and cleaning up the resources in the
    pool, respectively. Claiming a resource of the pool for a block of
    code is done using a with statement on the transaction method. The
    transaction method also allows filtering of the pool and supplying
    a default value to be used as the resource if no resources are
    free.

    Example::

        from riak.Pool import Pool, BadResource
        class ListPool(Pool):
            def create_resource(self):
                return []

            def destroy_resource(self):
                # Lists don't need to be cleaned up
                pass

        pool = ListPool()
        with pool.transaction() as resource:
            resource.append(1)
        with pool.transaction() as resource2:
            print(repr(resource2)) # should be [1]

    """

    def __init__(self):
        """
        Creates a new Pool. This should be called manually if you
        override the :meth:`__init__` method in a subclass.
        """
        self.lock = threading.RLock()
        self.releaser = threading.Condition(self.lock)
        self.resources = list()

    def acquire(self, _filter=None, default=None):
        """
        acquire(_filter=None, default=None)

        Claims a resource from the pool for manual use. Resources are
        created as needed when all members of the pool are claimed or
        the pool is empty. Most of the time you will want to use
        :meth:`transaction`.

        :param _filter: a filter that can be used to select a member
            of the pool
        :type _filter: callable
        :param default: a value that will be used instead of calling
            :meth:`create_resource` if a new resource needs to be created
        :rtype: Resource
        """
        if not _filter:
            def _filter(obj):
                return True
        elif not callable(_filter):
            raise TypeError("_filter is not a callable")

        resource = None
        with self.lock:
            for e in self.resources:
                if not e.claimed and _filter(e.object):
                    resource = e
                    break
            if resource is None:
                if default is not None:
                    resource = Resource(default, self)
                else:
                    resource = Resource(self.create_resource(), self)
                self.resources.append(resource)
            resource.claimed = True
        return resource

    def release(self, resource):
        """release(resource)

        Returns a resource to the pool. Most of the time you will want
        to use :meth:`transaction`, but if you use :meth:`acquire`,
        you must release the acquired resource back to the pool when
        finished. Failure to do so could result in deadlock.

        :param resource: Resource
        """
        with self.releaser:
            resource.claimed = False
            self.releaser.notify_all()

    @contextmanager
    def transaction(self, _filter=None, default=None):
        """
        transaction(_filter=None, default=None)

        Claims a resource from the pool for use in a thread-safe,
        reentrant manner (as part of a with statement). Resources are
        created as needed when all members of the pool are claimed or
        the pool is empty.

        :param _filter: a filter that can be used to select a member
            of the pool
        :type _filter: callable
        :param default: a value that will be used instead of calling
            :meth:`create_resource` if a new resource needs to be created
        """
        resource = self.acquire(_filter=_filter, default=default)
        try:
            yield resource.object
        except BadResource:
            self.delete_resource(resource)
            raise
        finally:
            self.release(resource)

    def delete_resource(self, resource):
        """
        Deletes the resource from the pool and destroys the associated
        resource. Not usually needed by users of the pool, but called
        internally when BadResource is raised.

        :param resource: the resource to remove
        :type resource: Resource
        """
        with self.lock:
            self.resources.remove(resource)
        self.destroy_resource(resource.object)
        del resource

    def __iter__(self):
        """
        Iterator callback to iterate over the resources of the pool.
        """
        return PoolIterator(self)

    def clear(self):
        """
        Removes all resources from the pool, calling :meth:`delete_resource`
        with each one so that the resources are cleaned up.
        """
        for resource in self:
            self.delete_resource(resource)

    def create_resource(self):
        """
        Implemented by subclasses to allocate a new resource for use
        in the pool.
        """
        raise NotImplementedError

    def destroy_resource(self, obj):
        """
        Called when removing a resource from the pool so that it can
        be cleanly deallocated. Subclasses should implement this
        method if additional cleanup is needed beyond normal GC. The
        default implementation is a no-op.

        :param obj: the resource being removed
        """
        pass


class PoolIterator(object):
    """
    Iterates over a snapshot of the pool in a thread-safe manner,
    eventually touching all resources that were known when the
    iteration started.

    Note that if claimed resources are not released for long periods,
    the iterator may hang, waiting for those last resources to be
    released. The iteration and pool functionality is only meant to be
    used internally within the client, and resources will be claimed
    per client operation, making this an unlikely event (although
    still possible).
    """

    def __init__(self, pool):
        with pool.lock:
            self.targets = pool.resources[:]
        self.unlocked = []
        self.lock = pool.lock
        self.releaser = pool.releaser

    def __iter__(self):
        return self

    def next(self):
        # Python 2.x version
        if len(self.targets) == 0:
            raise StopIteration
        if len(self.unlocked) == 0:
            self.__claim_resources()
        return self.unlocked.pop(0)

    def __next__(self):
        # Python 3.x version
        return self.next()

    def __claim_resources(self):
        with self.lock:
            with self.releaser:
                if self.__all_claimed():
                    self.releaser.wait()
            for resource in self.targets:
                if not resource.claimed:
                    self.targets.remove(resource)
                    self.unlocked.append(resource)
                    resource.claimed = True

    def __all_claimed(self):
        for resource in self.targets:
            if not resource.claimed:
                return False
        return True



"""
Copyright 2015 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from collections import Sized
from riak.datatypes.datatype import Datatype
from six import string_types
from riak.datatypes import TYPES


class Register(Sized, Datatype):
    """
    A convergent datatype that represents an opaque string that is set
    with last-write-wins semantics, and may only be embedded in
    :class:`~riak.datatypes.Map` instances.
    """

    type_name = 'register'
    _type_error_msg = "Registers can only be strings"

    def _post_init(self):
        self._new_value = None

    def _default_value(self):
        return ""

    @Datatype.value.getter
    def value(self):
        """
        Returns a copy of the original value of the register.

        :rtype: str
        """
        return self._value[:]

    @Datatype.modified.getter
    def modified(self):
        """
        Whether this register has staged assignment.
        """
        return self._new_value is not None

    def to_op(self):
        """
        Extracts the mutation operation from the register.

        :rtype: str, None
        """
        if self._new_value is not None:
            return ('assign', self._new_value)

    def assign(self, new_value):
        """
        Assigns a new value to the register.

        :param new_value: the new value for the register
        :type new_value: str
        """
        self._raise_if_badtype(new_value)
        self._new_value = new_value

    def __len__(self):
        return len(self.value)

    def _check_type(self, new_value):
        return isinstance(new_value, string_types)


TYPES['register'] = Register



"""
Copyright 2015 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import socket
import select
from six import PY2
from riak.security import SecurityError, USE_STDLIB_SSL
from riak.transports.pool import Pool
from riak.transports.http.transport import RiakHttpTransport
if USE_STDLIB_SSL:
    import ssl
    from riak.transports.security import configure_ssl_context
else:
    import OpenSSL.SSL
    from riak.transports.security import RiakWrappedSocket,\
        configure_pyopenssl_context
if PY2:
    from httplib import HTTPConnection, \
        NotConnected, \
        IncompleteRead, \
        ImproperConnectionState, \
        BadStatusLine, \
        HTTPSConnection
else:
    from http.client import HTTPConnection, \
        HTTPSConnection, \
        NotConnected, \
        IncompleteRead, \
        ImproperConnectionState, \
        BadStatusLine


class NoNagleHTTPConnection(HTTPConnection):
    """
    Setup a connection class which does not use Nagle - deal with
    latency on PUT requests lower than MTU
    """
    def connect(self):
        """
        Set TCP_NODELAY on socket
        """
        HTTPConnection.connect(self)
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)


# Inspired by
# http://code.activestate.com/recipes/577548-https-httplib-client-connection-with-certificate-v/
class RiakHTTPSConnection(HTTPSConnection):
    def __init__(self,
                 host,
                 port,
                 credentials,
                 pkey_file=None,
                 cert_file=None,
                 timeout=None):
        """
        Class to make a HTTPS connection,
        with support for full client-based SSL Authentication

        :param host: Riak host name
        :type host: str
        :param port: Riak host port number
        :type port: int
        :param credentials: Security Credential settings
        :type  credentials: SecurityCreds
        :param pkey_file: PEM formatted file that contains your private key
        :type pkey_file: str
        :param cert_file: PEM formatted certificate chain file
        :type cert_file: str
        :param timeout: Number of seconds before timing out
        :type timeout: int
        """
        if PY2:
            # NB: it appears that pkey_file / cert_file are never set
            # in riak/transports/http/connection.py#_connect() method
            pkf = pkey_file
            if pkf is None and credentials is not None:
                pkf = credentials._pkey_file

            cf = cert_file
            if cf is None and credentials is not None:
                cf = credentials._cert_file

            HTTPSConnection.__init__(self,
                                     host,
                                     port,
                                     key_file=pkf,
                                     cert_file=cf)
        else:
            super(RiakHTTPSConnection, self). \
                __init__(host=host,
                         port=port,
                         key_file=credentials._pkey_file,
                         cert_file=credentials._cert_file)
        self.pkey_file = pkey_file
        self.cert_file = cert_file
        self.credentials = credentials
        self.timeout = timeout

    def connect(self):
        """
        Connect to a host on a given (SSL) port using PyOpenSSL.
        """
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if not USE_STDLIB_SSL:
            ssl_ctx = configure_pyopenssl_context(self.credentials)

            # attempt to upgrade the socket to TLS
            cxn = OpenSSL.SSL.Connection(ssl_ctx, sock)
            cxn.set_connect_state()
            while True:
                try:
                    cxn.do_handshake()
                except OpenSSL.SSL.WantReadError:
                    select.select([sock], [], [])
                    continue
                except OpenSSL.SSL.Error as e:
                    raise SecurityError('bad handshake - ' + str(e))
                break

            self.sock = RiakWrappedSocket(cxn, sock)
            self.credentials._check_revoked_cert(self.sock)
        else:
            ssl_ctx = configure_ssl_context(self.credentials)
            if self.timeout is not None:
                sock.settimeout(self.timeout)
            self.sock = ssl.SSLSocket(sock=sock,
                                      keyfile=self.credentials.pkey_file,
                                      certfile=self.credentials.cert_file,
                                      cert_reqs=ssl.CERT_REQUIRED,
                                      ca_certs=self.credentials.cacert_file,
                                      ciphers=self.credentials.ciphers,
                                      server_hostname=self.host)
            self.sock.context = ssl_ctx


class RiakHttpPool(Pool):
    """
    A pool of HTTP(S) transport connections.
    """
    def __init__(self, client, **options):
        self.client = client
        self.options = options
        self.connection_class = NoNagleHTTPConnection
        if self.client._credentials:
            self.connection_class = RiakHTTPSConnection

        super(RiakHttpPool, self).__init__()

    def create_resource(self):
        node = self.client._choose_node()
        return RiakHttpTransport(node=node,
                                 client=self.client,
                                 connection_class=self.connection_class,
                                 **self.options)

    def destroy_resource(self, transport):
        transport.close()


CONN_CLOSED_ERRORS = (
    NotConnected,
    IncompleteRead,
    ImproperConnectionState,
    BadStatusLine
)


def is_retryable(err):
    """
    Determines if the given exception is something that is
    network/socket-related and should thus cause the HTTP connection
    to close and the operation retried on another node.

    :rtype: boolean
    """
    for errtype in CONN_CLOSED_ERRORS:
        if isinstance(err, errtype):
            return True
    return False
